# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for prepare utility functions."""

import logging
import os
import unittest

from lsst.ctrl.bps import (
    BPS_DEFAULTS,
    BPS_SEARCH_ORDER,
    BpsConfig,
    GenericWorkflowExec,
    GenericWorkflowFile,
    GenericWorkflowJob,
)
from lsst.ctrl.bps.htcondor import prepare_utils
from lsst.ctrl.bps.tests.gw_test_utils import make_3_label_workflow, make_3_label_workflow_groups_sort

logger = logging.getLogger("lsst.ctrl.bps.htcondor")

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class TranslateJobCmdsTestCase(unittest.TestCase):
    """Test _translate_job_cmds method."""

    def setUp(self):
        self.gw_exec = GenericWorkflowExec("test_exec", "/dummy/dir/pipetask")
        self.cached_vals = {"profile": {}, "bpsUseShared": True, "memoryLimit": 32768}

    def testRetryUnlessNone(self):
        gwjob = GenericWorkflowJob("retryUnless", "label1", executable=self.gw_exec)
        gwjob.retry_unless_exit = None
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertNotIn("retry_until", htc_commands)

    def testRetryUnlessInt(self):
        gwjob = GenericWorkflowJob("retryUnlessInt", "label1", executable=self.gw_exec)
        gwjob.retry_unless_exit = 3
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(int(htc_commands["retry_until"]), gwjob.retry_unless_exit)

    def testRetryUnlessList(self):
        gwjob = GenericWorkflowJob("retryUnlessList", "label1", executable=self.gw_exec)
        gwjob.retry_unless_exit = [1, 2]
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["retry_until"], "member(ExitCode, {1,2})")

    def testRetryUnlessBad(self):
        gwjob = GenericWorkflowJob("retryUnlessBad", "label1", executable=self.gw_exec)
        gwjob.retry_unless_exit = "1,2,3"
        with self.assertRaises(ValueError) as cm:
            _ = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertIn("retryUnlessExit", str(cm.exception))

    def testEnvironmentBasic(self):
        gwjob = GenericWorkflowJob("jobEnvironment", "label1", executable=self.gw_exec)
        gwjob.environment = {"TEST_INT": 1, "TEST_STR": "TWO"}
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["environment"], "TEST_INT=1 TEST_STR='TWO'")

    def testEnvironmentSpaces(self):
        gwjob = GenericWorkflowJob("jobEnvironment", "label1", executable=self.gw_exec)
        gwjob.environment = {"TEST_SPACES": "spacey value"}
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["environment"], "TEST_SPACES='spacey value'")

    def testEnvironmentSingleQuotes(self):
        gwjob = GenericWorkflowJob("jobEnvironment", "label1", executable=self.gw_exec)
        gwjob.environment = {"TEST_SINGLE_QUOTES": "spacey 'quoted' value"}
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["environment"], "TEST_SINGLE_QUOTES='spacey ''quoted'' value'")

    def testEnvironmentDoubleQuotes(self):
        gwjob = GenericWorkflowJob("jobEnvironment", "label1", executable=self.gw_exec)
        gwjob.environment = {"TEST_DOUBLE_QUOTES": 'spacey "double" value'}
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["environment"], """TEST_DOUBLE_QUOTES='spacey ""double"" value'""")

    def testEnvironmentWithEnvVars(self):
        gwjob = GenericWorkflowJob("jobEnvironment", "label1", executable=self.gw_exec)
        gwjob.environment = {"TEST_ENV_VAR": "<ENV:CTRL_BPS_DIR>/tests"}
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["environment"], "TEST_ENV_VAR='$ENV(CTRL_BPS_DIR)/tests'")

    def testPeriodicRelease(self):
        gwjob = GenericWorkflowJob("periodicRelease", "label1", executable=self.gw_exec)
        gwjob.request_memory = 2048
        gwjob.memory_multiplier = 2
        gwjob.number_of_retries = 3
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        release = (
            "JobStatus == 5 && NumJobStarts <= JobMaxRetries && "
            "(HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 || "
            "HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34) && "
            "min({int(2048 * pow(2, NumJobStarts - 1)), 32768}) < 32768"
        )
        self.assertEqual(htc_commands["periodic_release"], release)

    def testPeriodicRemoveNoRetries(self):
        gwjob = GenericWorkflowJob("periodicRelease", "label1", executable=self.gw_exec)
        gwjob.request_memory = 2048
        gwjob.memory_multiplier = 1
        gwjob.number_of_retries = 0
        htc_commands = prepare_utils._translate_job_cmds(self.cached_vals, None, gwjob)
        remove = "JobStatus == 5 && (NumJobStarts > JobMaxRetries)"
        self.assertEqual(htc_commands["periodic_remove"], remove)
        self.assertEqual(htc_commands["max_retries"], 0)


class GroupToSubdagTestCase(unittest.TestCase):
    """Test _group_to_subdag function."""

    def testBlocking(self):
        gw = make_3_label_workflow_groups_sort("test1", True)
        gwjob = gw.get_job("group_order1_10001")
        config = BpsConfig(
            {},
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
        )

        htc_job = prepare_utils._group_to_subdag(config, gwjob, "the_prefix")
        self.assertEqual(len(htc_job.subdag), len(gwjob))


class GatherSiteValuesTestCase(unittest.TestCase):
    """Test _gather_site_values function."""

    def testAllThere(self):
        config = BpsConfig(
            {},
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
        )
        compute_site = "notThere"
        results = prepare_utils._gather_site_values(config, compute_site)
        self.assertEqual(results["memoryLimit"], BPS_DEFAULTS["memoryLimit"])

    def testNotSpecified(self):
        config = BpsConfig(
            {},
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        compute_site = "notThere"
        results = prepare_utils._gather_site_values(config, compute_site)
        self.assertEqual(results["memoryLimit"], BPS_DEFAULTS["memoryLimit"])


class GatherLabelValuesTestCase(unittest.TestCase):
    """Test _gather_labels_values function."""

    def testClusterLabel(self):
        # Test cluster value overrides pipetask.
        label = "label1"
        config = BpsConfig(
            {
                "cluster": {
                    "label1": {
                        "releaseExpr": "cluster_val",
                        "overwriteJobFiles": False,
                        "profile": {"condor": {"prof_val1": 3}},
                    }
                },
                "pipetask": {"label1": {"releaseExpr": "pipetask_val"}},
            },
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        results = prepare_utils._gather_label_values(config, label)
        self.assertEqual(
            results,
            {
                "attrs": {},
                "profile": {"prof_val1": 3},
                "releaseExpr": "cluster_val",
                "overwriteJobFiles": False,
            },
        )

    def testPipetaskLabel(self):
        label = "label1"
        config = BpsConfig(
            {
                "pipetask": {
                    "label1": {
                        "releaseExpr": "pipetask_val",
                        "overwriteJobFiles": False,
                        "profile": {"condor": {"prof_val1": 3}},
                    }
                }
            },
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        results = prepare_utils._gather_label_values(config, label)
        self.assertEqual(
            results,
            {
                "attrs": {},
                "profile": {"prof_val1": 3},
                "releaseExpr": "pipetask_val",
                "overwriteJobFiles": False,
            },
        )

    def testNoSection(self):
        label = "notThere"
        config = BpsConfig(
            {},
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        results = prepare_utils._gather_label_values(config, label)
        self.assertEqual(results, {"attrs": {}, "profile": {}, "overwriteJobFiles": True})

    def testNoOverwriteSpecified(self):
        label = "notthere"
        config = BpsConfig(
            {},
            search_order=BPS_SEARCH_ORDER,
            defaults={},
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        results = prepare_utils._gather_label_values(config, label)
        self.assertEqual(results, {"attrs": {}, "profile": {}, "overwriteJobFiles": True})

    def testFinalJob(self):
        label = "finalJob"
        config = BpsConfig(
            {"finalJob": {"profile": {"condor": {"prof_val2": 6, "+attr_val1": 5}}}},
            search_order=BPS_SEARCH_ORDER,
            defaults=BPS_DEFAULTS,
            wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService",
        )
        results = prepare_utils._gather_label_values(config, label)
        self.assertEqual(
            results, {"attrs": {"attr_val1": 5}, "profile": {"prof_val2": 6}, "overwriteJobFiles": False}
        )


class CreateCheckJobTestCase(unittest.TestCase):
    """Test _create_check_job function."""

    def testSuccess(self):
        group_job_name = "group_order1_val1a"
        job_label = "order1"
        job = prepare_utils._create_check_job(group_job_name, job_label)
        self.assertIn(group_job_name, job.name)
        self.assertEqual(job.label, job_label)
        self.assertIn("check_group_status.sub", job.subfile)


class CreatePeriodicReleaseExprTestCase(unittest.TestCase):
    """Test _create_periodic_release_expr function."""

    def testNoReleaseExpr(self):
        results = prepare_utils._create_periodic_release_expr(2048, 1, 32768, "")
        self.assertEqual(results, "")

    def testMultiplierNone(self):
        results = prepare_utils._create_periodic_release_expr(2048, None, 32768, "")
        self.assertEqual(results, "")

    def testJustMemoryReleaseExpr(self):
        self.maxDiff = None  # so test error shows entire strings
        results = prepare_utils._create_periodic_release_expr(2048, 2, 32768, "")
        truth = (
            "JobStatus == 5 && NumJobStarts <= JobMaxRetries && "
            "(HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 || "
            "HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34) && "
            "min({int(2048 * pow(2, NumJobStarts - 1)), 32768}) < 32768"
        )
        self.assertEqual(results, truth)

    def testJustUserReleaseExpr(self):
        results = prepare_utils._create_periodic_release_expr(2048, 1, 32768, "True")
        truth = "JobStatus == 5 && NumJobStarts <= JobMaxRetries && HoldReasonCode =!= 1 && True"
        self.assertEqual(results, truth)

    def testJustUserReleaseExprMultiplierNone(self):
        results = prepare_utils._create_periodic_release_expr(2048, None, 32768, "True")
        truth = "JobStatus == 5 && NumJobStarts <= JobMaxRetries && HoldReasonCode =!= 1 && True"
        self.assertEqual(results, truth)

    def testMemoryAndUserReleaseExpr(self):
        self.maxDiff = None  # so test error shows entire strings
        results = prepare_utils._create_periodic_release_expr(2048, 2, 32768, "True")
        truth = (
            "JobStatus == 5 && NumJobStarts <= JobMaxRetries && "
            "((HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 || "
            "HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34) && "
            "min({int(2048 * pow(2, NumJobStarts - 1)), 32768}) < 32768 || "
            "HoldReasonCode =!= 1 && True)"
        )
        self.assertEqual(results, truth)


class CreatePeriodicRemoveExprTestCase(unittest.TestCase):
    """Test _create_periodic_release_expr function."""

    def testBasicRemoveExpr(self):
        """Function assumes only called if max_retries >= 0."""
        results = prepare_utils._create_periodic_remove_expr(2048, 1, 32768)
        truth = "JobStatus == 5 && (NumJobStarts > JobMaxRetries)"
        self.assertEqual(results, truth)

    def testBasicRemoveExprMultiplierNone(self):
        """Function assumes only called if max_retries >= 0."""
        results = prepare_utils._create_periodic_remove_expr(2048, None, 32768)
        truth = "JobStatus == 5 && (NumJobStarts > JobMaxRetries)"
        self.assertEqual(results, truth)

    def testMemoryRemoveExpr(self):
        self.maxDiff = None  # so test error shows entire strings
        results = prepare_utils._create_periodic_remove_expr(2048, 2, 32768)
        truth = (
            "JobStatus == 5 && (NumJobStarts > JobMaxRetries || "
            "((HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 || "
            "HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34) && "
            "min({int(2048 * pow(2, NumJobStarts - 1)), 32768}) == 32768))"
        )
        self.assertEqual(results, truth)


class HandleJobOutputsTestCase(unittest.TestCase):
    """Test _handle_job_outputs function."""

    def setUp(self):
        self.job_name = "test_job"
        self.out_prefix = "/test/prefix"

    def tearDown(self):
        pass

    def testNoOutputsSharedFilesystem(self):
        """Test with shared filesystem and no outputs."""
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = []

        result = prepare_utils._handle_job_outputs(mock_workflow, self.job_name, True, self.out_prefix)

        self.assertEqual(result, {"transfer_output_files": '""'})

    def testWithOutputsSharedFilesystem(self):
        """Test with shared filesystem and outputs present (still empty)."""
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = [
            GenericWorkflowFile(name="output.txt", src_uri="/path/to/output.txt")
        ]

        result = prepare_utils._handle_job_outputs(mock_workflow, self.job_name, True, self.out_prefix)

        self.assertEqual(result, {"transfer_output_files": '""'})

    def testNoOutputsNoSharedFilesystem(self):
        """Test without shared filesystem and no outputs."""
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = []

        result = prepare_utils._handle_job_outputs(mock_workflow, self.job_name, False, self.out_prefix)

        self.assertEqual(result, {"transfer_output_files": '""'})

    def testWithAnOutputNoSharedFilesystem(self):
        """Test without shared filesystem and single output file."""
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = [
            GenericWorkflowFile(name="output.txt", src_uri="/path/to/output.txt")
        ]

        result = prepare_utils._handle_job_outputs(mock_workflow, self.job_name, False, self.out_prefix)

        expected = {
            "transfer_output_files": "output.txt",
            "transfer_output_remaps": '"output.txt=/path/to/output.txt"',
        }
        self.assertEqual(result, expected)

    def testWithOutputsNoSharedFilesystem(self):
        """Test without shared filesystem and multiple output files."""
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = [
            GenericWorkflowFile(name="output1.txt", src_uri="/path/output1.txt"),
            GenericWorkflowFile(name="output2.txt", src_uri="/another/path/output2.txt"),
        ]

        result = prepare_utils._handle_job_outputs(mock_workflow, self.job_name, False, self.out_prefix)

        expected = {
            "transfer_output_files": "output1.txt,output2.txt",
            "transfer_output_remaps": '"output1.txt=/path/output1.txt;output2.txt=/another/path/output2.txt"',
        }
        self.assertEqual(result, expected)

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.prepare_utils._LOG")
    def testLogging(self, mock_log):
        mock_workflow = unittest.mock.Mock()
        mock_workflow.get_job_outputs.return_value = [
            GenericWorkflowFile(name="output.txt", src_uri="/path/to/output.txt")
        ]

        prepare_utils._handle_job_outputs(mock_workflow, self.job_name, False, self.out_prefix)

        self.assertTrue(mock_log.debug.called)
        debug_calls = mock_log.debug.call_args_list
        self.assertTrue(any("src_uri=" in str(call) for call in debug_calls))
        self.assertTrue(any("transfer_output_files=" in str(call) for call in debug_calls))
        self.assertTrue(any("transfer_output_remaps=" in str(call) for call in debug_calls))


class CreateJobTestCase(unittest.TestCase):
    """Test _create_job function."""

    def setUp(self):
        self.generic_workflow = make_3_label_workflow("test1", True)

    def testNoOverwrite(self):
        template = "{label}/{tract}/{patch}/{band}/{subfilter}/{physical_filter}/{visit}/{exposure}"
        cached_values = {
            "bpsUseShared": True,
            "overwriteJobFiles": False,
            "memoryLimit": 491520,
            "profile": {},
            "attrs": {},
        }
        gwjob = self.generic_workflow.get_final()
        out_prefix = "submit"
        htc_job = prepare_utils._create_job(template, cached_values, self.generic_workflow, gwjob, out_prefix)
        self.assertEqual(htc_job.name, gwjob.name)
        self.assertEqual(htc_job.label, gwjob.label)
        self.assertIn("NumJobStarts", htc_job.cmds["output"])
        self.assertIn("NumJobStarts", htc_job.cmds["error"])
        self.assertNotIn("NumJobStarts", htc_job.cmds["log"])
        self.assertTrue(htc_job.cmds["error"].endswith(".out"))
        self.assertTrue(htc_job.cmds["output"].endswith(".out"))
        self.assertTrue(htc_job.cmds["log"].endswith(".log"))


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utility functions for preparing the HTCondor workflow."""

import logging
import os
import re
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Any, cast

from lsst.ctrl.bps import (
    BpsConfig,
    GenericWorkflow,
    GenericWorkflowGroup,
    GenericWorkflowJob,
    GenericWorkflowNodeType,
    GenericWorkflowNoopJob,
)
from lsst.ctrl.bps.bps_utils import create_count_summary

from .lssthtc import (
    HTCDag,
    HTCJob,
    condor_status,
    htc_escape,
)

_LOG = logging.getLogger(__name__)

DEFAULT_HTC_EXEC_PATT = ".*worker.*"
"""Default pattern for searching execute machines in an HTCondor pool.
"""


def _create_job(subdir_template, cached_values, generic_workflow, gwjob, out_prefix):
    """Convert GenericWorkflow job nodes to DAG jobs.

    Parameters
    ----------
    subdir_template : `str`
        Template for making subdirs.
    cached_values : `dict`
        Site and label specific values.
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        Generic workflow that is being converted.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        The generic job to convert to a HTCondor job.
    out_prefix : `str`
        Directory prefix for HTCondor files.

    Returns
    -------
    htc_job : `lsst.ctrl.bps.wms.htcondor.HTCJob`
        The HTCondor job equivalent to the given generic job.
    """
    htc_job = HTCJob(gwjob.name, label=gwjob.label)

    curvals = defaultdict(str)
    curvals["label"] = gwjob.label
    if gwjob.tags:
        curvals.update(gwjob.tags)

    subdir = Path("jobs") / subdir_template.format_map(curvals)
    htc_job.subdir = subdir
    htc_job.subfile = f"{gwjob.name}.sub"
    htc_job.add_dag_cmds({"dir": subdir})

    htc_job_cmds = {
        "universe": "vanilla",
        "should_transfer_files": "YES",
        "when_to_transfer_output": "ON_EXIT_OR_EVICT",
        "transfer_output_files": '""',  # Set to empty string to disable
        "transfer_executable": "False",
        "getenv": "True",
        # Exceeding memory sometimes triggers SIGBUS or SIGSEGV error. Tell
        # htcondor to put on hold any jobs which exited by a signal. If
        # executed in a bash script, like finalJob, the signals will become
        # exit codes above 128 (exit code = 128 + signal number).
        "on_exit_hold": "ExitBySignal == true || ExitCode > 128",
        "on_exit_hold_reason": "ExitBySignal == true ? "
        'strcat("Job raised a signal ", string(ExitSignal), '
        '". Handling job as if it has gone over memory limit.") : '
        'strcat("Job exit code (", string(ExitCode), ") > 128. '
        'Handling job as if it has gone over memory limit.")',
        "on_exit_hold_subcode": "34",
    }

    htc_job_cmds.update(_translate_job_cmds(cached_values, generic_workflow, gwjob))

    # Combine stdout and stderr to reduce the number of files.
    for key in ("output", "error"):
        if cached_values["overwriteJobFiles"]:
            htc_job_cmds[key] = f"{gwjob.name}.$(Cluster).out"
        else:
            htc_job_cmds[key] = f"{gwjob.name}.$(Cluster).$$([NumJobStarts ?: 0]).out"
        _LOG.debug("HTCondor %s = %s", key, htc_job_cmds[key])

    key = "log"
    htc_job_cmds[key] = f"{gwjob.name}.$(Cluster).{key}"
    _LOG.debug("HTCondor %s = %s", key, htc_job_cmds[key])

    htc_job_cmds.update(
        _handle_job_inputs(generic_workflow, gwjob.name, cached_values["bpsUseShared"], out_prefix)
    )

    htc_job_cmds.update(
        _handle_job_outputs(generic_workflow, gwjob.name, cached_values["bpsUseShared"], out_prefix)
    )

    # Add the job cmds dict to the job object.
    htc_job.add_job_cmds(htc_job_cmds)

    htc_job.add_dag_cmds(_translate_dag_cmds(gwjob))

    # Add job attributes to job.
    _LOG.debug("gwjob.attrs = %s", gwjob.attrs)
    htc_job.add_job_attrs(gwjob.attrs)
    htc_job.add_job_attrs(cached_values["attrs"])
    htc_job.add_job_attrs({"bps_job_quanta": create_count_summary(gwjob.quanta_counts)})
    htc_job.add_job_attrs({"bps_job_name": gwjob.name, "bps_job_label": gwjob.label})

    return htc_job


def _translate_job_cmds(cached_vals, generic_workflow, gwjob):
    """Translate the job data that are one to one mapping

    Parameters
    ----------
    cached_vals : `dict` [`str`, `~typing.Any`]
        Config values common to jobs with same site or label.
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        Generic workflow that contains job to being converted.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        Generic workflow job to be converted.

    Returns
    -------
    htc_job_commands : `dict` [`str`, `~typing.Any`]
        Contains commands which can appear in the HTCondor submit description
        file.
    """
    # Values in the job script that just are name mappings.
    job_translation = {
        "mail_to": "notify_user",
        "when_to_mail": "notification",
        "request_cpus": "request_cpus",
        "priority": "priority",
        "category": "category",
        "accounting_group": "accounting_group",
        "accounting_user": "accounting_group_user",
    }

    jobcmds = {}
    for gwkey, htckey in job_translation.items():
        jobcmds[htckey] = getattr(gwjob, gwkey, None)

    # If accounting info was not set explicitly, use site settings if any.
    if not gwjob.accounting_group:
        jobcmds["accounting_group"] = cached_vals.get("accountingGroup")
    if not gwjob.accounting_user:
        jobcmds["accounting_group_user"] = cached_vals.get("accountingUser")

    # job commands that need modification
    if gwjob.retry_unless_exit:
        if isinstance(gwjob.retry_unless_exit, int):
            jobcmds["retry_until"] = f"{gwjob.retry_unless_exit}"
        elif isinstance(gwjob.retry_unless_exit, list):
            jobcmds["retry_until"] = (
                f"member(ExitCode, {{{','.join([str(x) for x in gwjob.retry_unless_exit])}}})"
            )
        else:
            raise ValueError("retryUnlessExit must be an integer or a list of integers.")

    if gwjob.request_disk:
        jobcmds["request_disk"] = f"{gwjob.request_disk}MB"

    if gwjob.request_memory:
        jobcmds["request_memory"] = f"{gwjob.request_memory}"

    memory_max = 0
    if gwjob.memory_multiplier:
        # Do not use try-except! At the moment, BpsConfig returns an empty
        # string if it does not contain the key.
        memory_limit = cached_vals["memoryLimit"]
        if not memory_limit:
            raise RuntimeError(
                "Memory autoscaling enabled, but automatic detection of the memory limit "
                "failed; setting it explicitly with 'memoryLimit' or changing worker node "
                "search pattern 'executeMachinesPattern' might help."
            )

        # Set maximal amount of memory job can ask for.
        #
        # The check below assumes that 'memory_limit' was set to a value which
        # realistically reflects actual physical limitations of a given compute
        # resource.
        memory_max = memory_limit
        if gwjob.request_memory_max and gwjob.request_memory_max < memory_limit:
            memory_max = gwjob.request_memory_max

        # Make job ask for more memory each time it failed due to insufficient
        # memory requirements.
        jobcmds["request_memory"] = _create_request_memory_expr(
            gwjob.request_memory, gwjob.memory_multiplier, memory_max
        )

    user_release_expr = cached_vals.get("releaseExpr", "")
    if gwjob.number_of_retries is not None and gwjob.number_of_retries >= 0:
        jobcmds["max_retries"] = gwjob.number_of_retries

        # No point in adding periodic_release if 0 retries
        if gwjob.number_of_retries > 0:
            periodic_release = _create_periodic_release_expr(
                gwjob.request_memory, gwjob.memory_multiplier, memory_max, user_release_expr
            )
            if periodic_release:
                jobcmds["periodic_release"] = periodic_release

        jobcmds["periodic_remove"] = _create_periodic_remove_expr(
            gwjob.request_memory, gwjob.memory_multiplier, memory_max
        )

    # Assume concurrency_limit implemented using HTCondor concurrency limits.
    # May need to move to special site-specific implementation if sites use
    # other mechanisms.
    if gwjob.concurrency_limit:
        jobcmds["concurrency_limit"] = gwjob.concurrency_limit

    # Handle command line
    if gwjob.executable.transfer_executable:
        jobcmds["transfer_executable"] = "True"
        jobcmds["executable"] = gwjob.executable.src_uri
    else:
        jobcmds["executable"] = _fix_env_var_syntax(gwjob.executable.src_uri)

    if gwjob.arguments:
        arguments = gwjob.arguments
        arguments = _replace_cmd_vars(arguments, gwjob)
        arguments = _replace_file_vars(cached_vals["bpsUseShared"], arguments, generic_workflow, gwjob)
        arguments = _fix_env_var_syntax(arguments)
        jobcmds["arguments"] = arguments

    if gwjob.environment:
        env_str = ""
        for name, value in gwjob.environment.items():
            if isinstance(value, str):
                value2 = _replace_cmd_vars(value, gwjob)
                value2 = _fix_env_var_syntax(value2)
                value2 = htc_escape(value2)
                env_str += f"{name}='{value2}' "  # Add single quotes to allow internal spaces
            else:
                env_str += f"{name}={value} "

        # Process above added one trailing space
        jobcmds["environment"] = env_str.rstrip()

    # Add extra "pass-thru" job commands
    if gwjob.profile:
        for key, val in gwjob.profile.items():
            jobcmds[key] = htc_escape(val)
    for key, val in cached_vals["profile"].items():
        jobcmds[key] = htc_escape(val)

    return jobcmds


def _translate_dag_cmds(gwjob):
    """Translate job values into DAGMan commands.

    Parameters
    ----------
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        Job containing values to be translated.

    Returns
    -------
    dagcmds : `dict` [`str`, `~typing.Any`]
        DAGMan commands for the job.
    """
    # Values in the dag script that just are name mappings.
    dag_translation = {"abort_on_value": "abort_dag_on", "abort_return_value": "abort_exit"}

    dagcmds = {}
    for gwkey, htckey in dag_translation.items():
        dagcmds[htckey] = getattr(gwjob, gwkey, None)

    # Still to be coded: vars "pre_cmdline", "post_cmdline"
    return dagcmds


def _fix_env_var_syntax(oldstr):
    """Change ENV place holders to HTCondor Env var syntax.

    Parameters
    ----------
    oldstr : `str`
        String in which environment variable syntax is to be fixed.

    Returns
    -------
    newstr : `str`
        Given string with environment variable syntax fixed.
    """
    newstr = oldstr
    for key in re.findall(r"<ENV:([^>]+)>", oldstr):
        newstr = newstr.replace(rf"<ENV:{key}>", f"$ENV({key})")
    return newstr


def _replace_file_vars(use_shared, arguments, workflow, gwjob):
    """Replace file placeholders in command line arguments with correct
    physical file names.

    Parameters
    ----------
    use_shared : `bool`
        Whether HTCondor can assume shared filesystem.
    arguments : `str`
        Arguments string in which to replace file placeholders.
    workflow : `lsst.ctrl.bps.GenericWorkflow`
        Generic workflow that contains file information.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        The job corresponding to the arguments.

    Returns
    -------
    arguments : `str`
        Given arguments string with file placeholders replaced.
    """
    # Replace input file placeholders with paths.
    for gwfile in workflow.get_job_inputs(gwjob.name, data=True, transfer_only=False):
        if not gwfile.wms_transfer:
            # Must assume full URI if in command line and told WMS is not
            # responsible for transferring file.
            uri = gwfile.src_uri
        elif use_shared:
            if gwfile.job_shared:
                # Have shared filesystems and jobs can share file.
                uri = gwfile.src_uri
            else:
                uri = os.path.basename(gwfile.src_uri)
        else:  # Using push transfer
            uri = os.path.basename(gwfile.src_uri)
        arguments = arguments.replace(f"<FILE:{gwfile.name}>", uri)

    # Replace output file placeholders with paths.
    for gwfile in workflow.get_job_outputs(gwjob.name, data=True, transfer_only=False):
        if not gwfile.wms_transfer:
            # Must assume full URI if in command line and told WMS is not
            # responsible for transferring file.
            uri = gwfile.src_uri
        elif use_shared:
            if gwfile.job_shared:
                # Have shared filesystems and jobs can share file.
                uri = gwfile.src_uri
            else:
                uri = os.path.basename(gwfile.src_uri)
        else:  # Using push transfer
            uri = os.path.basename(gwfile.src_uri)
        arguments = arguments.replace(f"<FILE:{gwfile.name}>", uri)
    return arguments


def _replace_cmd_vars(arguments, gwjob):
    """Replace format-style placeholders in arguments.

    Parameters
    ----------
    arguments : `str`
        Arguments string in which to replace placeholders.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        Job containing values to be used to replace placeholders
        (in particular gwjob.cmdvals).

    Returns
    -------
    arguments : `str`
        Given arguments string with placeholders replaced.
    """
    replacements = gwjob.cmdvals if gwjob.cmdvals is not None else {}
    try:
        arguments = arguments.format(**replacements)
    except (KeyError, TypeError) as exc:  # TypeError in case None instead of {}
        _LOG.error("Could not replace command variables: replacement for %s not provided", str(exc))
        _LOG.debug("arguments: %s\ncmdvals: %s", arguments, replacements)
        raise
    return arguments


def _handle_job_inputs(
    generic_workflow: GenericWorkflow, job_name: str, use_shared: bool, out_prefix: str
) -> dict[str, str]:
    """Add job input files from generic workflow to job.

    Parameters
    ----------
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        The generic workflow (e.g., has executable name and arguments).
    job_name : `str`
        Unique name for the job.
    use_shared : `bool`
        Whether job has access to files via shared filesystem.
    out_prefix : `str`
        The root directory into which all WMS-specific files are written.

    Returns
    -------
    htc_commands : `dict` [`str`, `str`]
        HTCondor commands for the job submission script.
    """
    inputs = []
    for gwf_file in generic_workflow.get_job_inputs(job_name, data=True, transfer_only=True):
        _LOG.debug("src_uri=%s", gwf_file.src_uri)

        uri = Path(gwf_file.src_uri)

        # Note if use_shared and job_shared, don't need to transfer file.

        if not use_shared:  # Copy file using push to job
            inputs.append(str(uri))
        elif not gwf_file.job_shared:  # Jobs require own copy
            # if using shared filesystem, but still need copy in job. Use
            # HTCondor's curl plugin for a local copy.
            if uri.is_dir():
                raise RuntimeError(
                    f"HTCondor plugin cannot transfer directories locally within job {gwf_file.src_uri}"
                )
            inputs.append(f"file://{uri}")

    htc_commands = {}
    if inputs:
        htc_commands["transfer_input_files"] = ",".join(inputs)
        _LOG.debug("transfer_input_files=%s", htc_commands["transfer_input_files"])
    return htc_commands


def _handle_job_outputs(
    generic_workflow: GenericWorkflow, job_name: str, use_shared: bool, out_prefix: str
) -> dict[str, str]:
    """Add job output files from generic workflow to the job if any.

    Parameters
    ----------
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        The generic workflow (e.g., has executable name and arguments).
    job_name : `str`
        Unique name for the job.
    use_shared : `bool`
        Whether job has access to files via shared filesystem.
    out_prefix : `str`
        The root directory into which all WMS-specific files are written.

    Returns
    -------
    htc_commands : `dict` [`str`, `str`]
        HTCondor commands for the job submission script.
    """
    outputs = []
    output_remaps = []
    for gwf_file in generic_workflow.get_job_outputs(job_name, data=True, transfer_only=True):
        _LOG.debug("src_uri=%s", gwf_file.src_uri)

        uri = Path(gwf_file.src_uri)
        if not use_shared:
            outputs.append(uri.name)
            output_remaps.append(f"{uri.name}={str(uri)}")

    # Set to an empty string to disable and only update if there are output
    # files to transfer. Otherwise, HTCondor will transfer back all files in
    # the job’s temporary working directory that have been modified or created
    # by the job.
    htc_commands = {"transfer_output_files": '""'}
    if outputs:
        htc_commands["transfer_output_files"] = ",".join(outputs)
        _LOG.debug("transfer_output_files=%s", htc_commands["transfer_output_files"])

        htc_commands["transfer_output_remaps"] = f'"{";".join(output_remaps)}"'
        _LOG.debug("transfer_output_remaps=%s", htc_commands["transfer_output_remaps"])
    return htc_commands


def _create_periodic_release_expr(
    memory: int, multiplier: float | None, limit: int, additional_expr: str = ""
) -> str:
    """Construct an HTCondorAd expression for releasing held jobs.

    Parameters
    ----------
    memory : `int`
        Requested memory in MB.
    multiplier : `float` or None
        Memory growth rate between retries.
    limit : `int`
        Memory limit.
    additional_expr : `str`, optional
        Expression to add to periodic_release.  Defaults to empty string.

    Returns
    -------
    expr : `str`
        A string representing an HTCondor ClassAd expression for releasing job.
    """
    _LOG.debug(
        "periodic_release: memory: %s, multiplier: %s, limit: %s, additional_expr: %s",
        memory,
        multiplier,
        limit,
        additional_expr,
    )

    # ctrl_bps sets multiplier to None in the GenericWorkflow if
    # memoryMultiplier <= 1, but checking value just in case.
    if (not multiplier or multiplier <= 1) and not additional_expr:
        return ""

    # Job ClassAds attributes 'HoldReasonCode' and 'HoldReasonSubCode' are
    # UNDEFINED if job is not HELD (i.e. when 'JobStatus' is not 5).
    # The special comparison operators ensure that all comparisons below will
    # evaluate to FALSE in this case.
    #
    # Note:
    # May not be strictly necessary. Operators '&&' and '||' are not strict so
    # the entire expression should evaluate to FALSE when the job is not HELD.
    # According to ClassAd evaluation semantics FALSE && UNDEFINED is FALSE,
    # but better safe than sorry.
    is_held = "JobStatus == 5"
    is_retry_allowed = "NumJobStarts <= JobMaxRetries"

    mem_expr = ""
    if memory and multiplier and multiplier > 1 and limit:
        was_mem_exceeded = (
            "(HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 "
            "|| HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34)"
        )
        was_below_limit = f"min({{int({memory} * pow({multiplier}, NumJobStarts - 1)), {limit}}}) < {limit}"
        mem_expr = f"{was_mem_exceeded} && {was_below_limit}"

    user_expr = ""
    if additional_expr:
        # Never auto release a job held by user.
        user_expr = f"HoldReasonCode =!= 1 && {additional_expr}"

    expr = f"{is_held} && {is_retry_allowed}"
    if user_expr and mem_expr:
        expr += f" && ({mem_expr} || {user_expr})"
    elif user_expr:
        expr += f" && {user_expr}"
    elif mem_expr:
        expr += f" && {mem_expr}"

    return expr


def _create_periodic_remove_expr(memory, multiplier, limit):
    """Construct an HTCondorAd expression for removing jobs from the queue.

    Parameters
    ----------
    memory : `int`
        Requested memory in MB.
    multiplier : `float`
        Memory growth rate between retries.
    limit : `int`
        Memory limit.

    Returns
    -------
    expr : `str`
        A string representing an HTCondor ClassAd expression for removing jobs.
    """
    # Job ClassAds attributes 'HoldReasonCode' and 'HoldReasonSubCode'
    # are UNDEFINED if job is not HELD (i.e. when 'JobStatus' is not 5).
    # The special comparison operators ensure that all comparisons below
    # will evaluate to FALSE in this case.
    #
    # Note:
    # May not be strictly necessary. Operators '&&' and '||' are not
    # strict so the entire expression should evaluate to FALSE when the
    # job is not HELD. According to ClassAd evaluation semantics
    # FALSE && UNDEFINED is FALSE, but better safe than sorry.
    is_held = "JobStatus == 5"
    is_retry_disallowed = "NumJobStarts > JobMaxRetries"

    mem_expr = ""
    if memory and multiplier and multiplier > 1 and limit:
        mem_limit_expr = f"min({{int({memory} * pow({multiplier}, NumJobStarts - 1)), {limit}}}) == {limit}"

        mem_expr = (  # Add || here so only added if adding memory expr
            " || ((HoldReasonCode =?= 34 && HoldReasonSubCode =?= 0 "
            f"|| HoldReasonCode =?= 3 && HoldReasonSubCode =?= 34) && {mem_limit_expr})"
        )

    expr = f"{is_held} && ({is_retry_disallowed}{mem_expr})"
    return expr


def _create_request_memory_expr(memory, multiplier, limit):
    """Construct an HTCondor ClassAd expression for safe memory scaling.

    Parameters
    ----------
    memory : `int`
        Requested memory in MB.
    multiplier : `float`
        Memory growth rate between retries.
    limit : `int`
        Memory limit.

    Returns
    -------
    expr : `str`
        A string representing an HTCondor ClassAd expression enabling safe
        memory scaling between job retries.
    """
    # The check if the job was held due to exceeding memory requirements
    # will be made *after* job was released back to the job queue (is in
    # the IDLE state), hence the need to use `Last*` job ClassAds instead of
    # the ones describing job's current state.
    #
    # Also, 'Last*' job ClassAds attributes are UNDEFINED when a job is
    # initially put in the job queue. The special comparison operators ensure
    # that all comparisons below will evaluate to FALSE in this case.
    was_mem_exceeded = (
        "LastJobStatus =?= 5 "
        "&& (LastHoldReasonCode =?= 34 && LastHoldReasonSubCode =?= 0 "
        "|| LastHoldReasonCode =?= 3 && LastHoldReasonSubCode =?= 34)"
    )

    # If job runs the first time or was held for reasons other than exceeding
    # the memory, set the required memory to the requested value or use
    # the memory value measured by HTCondor (MemoryUsage) depending on
    # whichever is greater.
    expr = (
        f"({was_mem_exceeded}) "
        f"? min({{int({memory} * pow({multiplier}, NumJobStarts)), {limit}}}) "
        f": max({{{memory}, MemoryUsage ?: 0}})"
    )
    return expr


def _gather_site_values(config, compute_site):
    """Gather values specific to given site.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration that includes necessary submit/runtime
        information.
    compute_site : `str`
        Compute site name.

    Returns
    -------
    site_values : `dict` [`str`, `~typing.Any`]
        Values specific to the given site.
    """
    site_values = {"attrs": {}, "profile": {}}
    search_opts = {}
    if compute_site:
        search_opts["curvals"] = {"curr_site": compute_site}

    # Determine the hard limit for the memory requirement.
    found, limit = config.search("memoryLimit", opt=search_opts)
    if not found:
        search_opts["default"] = DEFAULT_HTC_EXEC_PATT
        _, patt = config.search("executeMachinesPattern", opt=search_opts)
        del search_opts["default"]

        # To reduce the amount of data, ignore dynamic slots (if any) as,
        # by definition, they cannot have more memory than
        # the partitionable slot they are the part of.
        constraint = f'SlotType != "Dynamic" && regexp("{patt}", Machine)'
        pool_info = condor_status(constraint=constraint)
        try:
            limit = max(int(info["TotalSlotMemory"]) for info in pool_info.values())
        except ValueError:
            _LOG.debug("No execute machine in the pool matches %s", patt)
    if limit:
        config[".bps_defined.memory_limit"] = limit

    _, site_values["bpsUseShared"] = config.search("bpsUseShared", opt={"default": False})
    site_values["memoryLimit"] = limit

    found, value = config.search("accountingGroup", opt=search_opts)
    if found:
        site_values["accountingGroup"] = value
    found, value = config.search("accountingUser", opt=search_opts)
    if found:
        site_values["accountingUser"] = value

    key = f".site.{compute_site}.profile.condor"
    if key in config:
        for subkey, val in config[key].items():
            if subkey.startswith("+"):
                site_values["attrs"][subkey[1:]] = val
            else:
                site_values["profile"][subkey] = val

    return site_values


def _gather_label_values(config: BpsConfig, label: str) -> dict[str, Any]:
    """Gather values specific to given job label.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration that includes necessary submit/runtime
        information.
    label : `str`
        GenericWorkflowJob label.

    Returns
    -------
    values : `dict` [`str`, `~typing.Any`]
        Values specific to the given job label.
    """
    values: dict[str, Any] = {"attrs": {}, "profile": {}}

    search_opts = {}
    profile_key = ""
    if label == "finalJob":
        search_opts["searchobj"] = config["finalJob"]
        profile_key = ".finalJob.profile.condor"
    elif label in config["cluster"]:
        search_opts["curvals"] = {"curr_cluster": label}
        profile_key = f".cluster.{label}.profile.condor"
    elif label in config["pipetask"]:
        search_opts["curvals"] = {"curr_pipetask": label}
        profile_key = f".pipetask.{label}.profile.condor"

    found, value = config.search("releaseExpr", opt=search_opts)
    if found:
        values["releaseExpr"] = value

    found, value = config.search("overwriteJobFiles", opt=search_opts)
    if found:
        values["overwriteJobFiles"] = value
    else:
        values["overwriteJobFiles"] = True

    if profile_key and profile_key in config:
        for subkey, val in config[profile_key].items():
            if subkey.startswith("+"):
                values["attrs"][subkey[1:]] = val
            else:
                values["profile"][subkey] = val

    return values


def _group_to_subdag(
    config: BpsConfig, generic_workflow_group: GenericWorkflowGroup, out_prefix: str
) -> HTCJob:
    """Convert a generic workflow group to an HTCondor dag.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        Workflow configuration.
    generic_workflow_group : `lsst.ctrl.bps.GenericWorkflowGroup`
        The generic workflow group to convert.
    out_prefix : `str`
        Location prefix to be used when creating jobs.

    Returns
    -------
    htc_job : `lsst.ctrl.bps.htcondor.HTCJob`
        Job for running the HTCondor dag.
    """
    jobname = f"wms_{generic_workflow_group.name}"
    htc_job = HTCJob(name=jobname, label=generic_workflow_group.label)
    htc_job.add_dag_cmds({"dir": f"subdags/{jobname}"})
    htc_job.subdag = _generic_workflow_to_htcondor_dag(config, generic_workflow_group, out_prefix)
    if not generic_workflow_group.blocking:
        htc_job.dagcmds["post"] = {
            "defer": "",
            "executable": f"{os.path.dirname(__file__)}/subdag_post.sh",
            "arguments": f"{jobname} $RETURN",
        }
    return htc_job


def _create_check_job(group_job_name: str, job_label: str) -> HTCJob:
    """Create a job to check status of a group job.

    Parameters
    ----------
    group_job_name : `str`
        Name of the group job.
    job_label : `str`
        Label to use for the check status job.

    Returns
    -------
    htc_job : `lsst.ctrl.bps.htcondor.HTCJob`
        Job description for the job to check group job status.
    """
    htc_job = HTCJob(name=f"wms_check_status_{group_job_name}", label=job_label)
    htc_job.subfile = "${CTRL_BPS_HTCONDOR_DIR}/python/lsst/ctrl/bps/htcondor/check_group_status.sub"
    htc_job.add_dag_cmds({"dir": f"subdags/{group_job_name}", "vars": {"group_job_name": group_job_name}})

    return htc_job


def _generic_workflow_to_htcondor_dag(
    config: BpsConfig, generic_workflow: GenericWorkflow, out_prefix: str
) -> HTCDag:
    """Convert a GenericWorkflow to a HTCDag.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        Workflow configuration.
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        The GenericWorkflow to convert.
    out_prefix : `str`
        Location prefix where the HTCondor files will be written.

    Returns
    -------
    dag : `lsst.ctrl.bps.htcondor.HTCDag`
        The HTCDag representation of the given GenericWorkflow.
    """
    dag = HTCDag(name=generic_workflow.name)

    _LOG.debug("htcondor dag attribs %s", generic_workflow.run_attrs)
    dag.add_attribs(generic_workflow.run_attrs)
    dag.add_attribs(
        {
            "bps_run_quanta": create_count_summary(generic_workflow.quanta_counts),
            "bps_job_summary": create_count_summary(generic_workflow.job_counts),
        }
    )

    _, tmp_template = config.search("subDirTemplate", opt={"replaceVars": False, "default": ""})
    if isinstance(tmp_template, str):
        subdir_template = defaultdict(lambda: tmp_template)
    else:
        subdir_template = tmp_template

    # Create all DAG jobs
    site_values = {}  # Cache compute site specific values to reduce config lookups.
    cached_values = {}  # Cache label-specific values to reduce config lookups.
    # Note: Can't use get_job_by_label because those only include payload jobs.
    for job_name in generic_workflow:
        gwjob = generic_workflow.get_job(job_name)
        if gwjob.node_type == GenericWorkflowNodeType.PAYLOAD:
            gwjob = cast(GenericWorkflowJob, gwjob)
            if gwjob.compute_site not in site_values:
                site_values[gwjob.compute_site] = _gather_site_values(config, gwjob.compute_site)
            if gwjob.label not in cached_values:
                cached_values[gwjob.label] = deepcopy(site_values[gwjob.compute_site])
                cached_values[gwjob.label].update(_gather_label_values(config, gwjob.label))
                _LOG.debug("cached: %s= %s", gwjob.label, cached_values[gwjob.label])
            htc_job = _create_job(
                subdir_template[gwjob.label],
                cached_values[gwjob.label],
                generic_workflow,
                gwjob,
                out_prefix,
            )
        elif gwjob.node_type == GenericWorkflowNodeType.NOOP:
            gwjob = cast(GenericWorkflowNoopJob, gwjob)
            htc_job = HTCJob(f"wms_{gwjob.name}", label=gwjob.label)
            htc_job.subfile = "${CTRL_BPS_HTCONDOR_DIR}/python/lsst/ctrl/bps/htcondor/noop.sub"
            htc_job.add_job_attrs({"bps_job_name": gwjob.name, "bps_job_label": gwjob.label})
            htc_job.add_dag_cmds({"noop": True})
        elif gwjob.node_type == GenericWorkflowNodeType.GROUP:
            gwjob = cast(GenericWorkflowGroup, gwjob)
            htc_job = _group_to_subdag(config, gwjob, out_prefix)
        else:
            raise RuntimeError(f"Unsupported generic workflow node type {gwjob.node_type} ({gwjob.name})")
        _LOG.debug("Calling adding job %s %s", htc_job.name, htc_job.label)
        dag.add_job(htc_job)

    # Add job dependencies to the DAG (be careful with wms_ jobs)
    for job_name in generic_workflow:
        gwjob = generic_workflow.get_job(job_name)
        parent_name = (
            gwjob.name if gwjob.node_type == GenericWorkflowNodeType.PAYLOAD else f"wms_{gwjob.name}"
        )
        successor_jobs = [generic_workflow.get_job(j) for j in generic_workflow.successors(job_name)]
        children_names = []
        if gwjob.node_type == GenericWorkflowNodeType.GROUP:
            gwjob = cast(GenericWorkflowGroup, gwjob)
            group_children = []  # Dependencies between same group jobs
            for sjob in successor_jobs:
                if sjob.node_type == GenericWorkflowNodeType.GROUP and sjob.label == gwjob.label:
                    group_children.append(f"wms_{sjob.name}")
                elif sjob.node_type == GenericWorkflowNodeType.PAYLOAD:
                    children_names.append(sjob.name)
                else:
                    children_names.append(f"wms_{sjob.name}")
            if group_children:
                dag.add_job_relationships([parent_name], group_children)
            if not gwjob.blocking:
                # Since subdag will always succeed, need to add a special
                # job that fails if group failed to block payload children.
                check_job = _create_check_job(f"wms_{gwjob.name}", gwjob.label)
                dag.add_job(check_job)
                dag.add_job_relationships([f"wms_{gwjob.name}"], [check_job.name])
                parent_name = check_job.name
        else:
            for sjob in successor_jobs:
                if sjob.node_type == GenericWorkflowNodeType.PAYLOAD:
                    children_names.append(sjob.name)
                else:
                    children_names.append(f"wms_{sjob.name}")

        dag.add_job_relationships([parent_name], children_names)

    # If final job exists in generic workflow, create DAG final job
    final = generic_workflow.get_final()
    if final and isinstance(final, GenericWorkflowJob):
        if final.compute_site and final.compute_site not in site_values:
            site_values[final.compute_site] = _gather_site_values(config, final.compute_site)
        if final.label not in cached_values:
            cached_values[final.label] = deepcopy(site_values[final.compute_site])
            cached_values[final.label].update(_gather_label_values(config, final.label))
        final_htjob = _create_job(
            subdir_template[final.label],
            cached_values[final.label],
            generic_workflow,
            final,
            out_prefix,
        )
        if "post" not in final_htjob.dagcmds:
            final_htjob.dagcmds["post"] = {
                "defer": "",
                "executable": f"{os.path.dirname(__file__)}/final_post.sh",
                "arguments": f"{final.name} $DAG_STATUS $RETURN",
            }
        dag.add_final_job(final_htjob)
    elif final and isinstance(final, GenericWorkflow):
        raise NotImplementedError("HTCondor plugin does not support a workflow as the final job")
    elif final:
        raise TypeError(f"Invalid type for GenericWorkflow.get_final() results ({type(final)})")

    return dag



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for DagmanConfigurator class."""

import logging
import os
import tempfile
import unittest
from pathlib import Path

from pydantic import ValidationError

from lsst.ctrl.bps import BpsConfig
from lsst.ctrl.bps.htcondor import HTCDag
from lsst.ctrl.bps.htcondor.dagman_configurator import DagmanConfigurator

logger = logging.getLogger("lsst.ctrl.bps.htcondor")


class DagmanConfiguratorTestCase(unittest.TestCase):
    """Unit tests for DagmanConfigurator class."""

    def setUp(self):
        self.config = BpsConfig(
            {
                "site": {
                    "foo": {
                        "wmsConfig": {"DAGMAN_USE_STRICT": 1},
                    },
                },
                "wmsConfig": {"DAGMAN_USE_STRICT": 0},
            }
        )

    def tearDown(self):
        pass

    def testInitDefaultSearchOptions(self):
        """Test object instantiation with default search options."""
        configurator = DagmanConfigurator(self.config)
        self.assertIn("DAGMAN_USE_STRICT", configurator.options)
        self.assertEqual(configurator.options["DAGMAN_USE_STRICT"], 0)
        self.assertIsNone(configurator.config_path)
        self.assertIsNone(configurator.prefix)

    def testInitCustomSearchOptions(self):
        """Test object instantiation with custom search options."""
        configurator = DagmanConfigurator(self.config, search_opts={"curvals": {"computeSite": "foo"}})
        self.assertIn("DAGMAN_USE_STRICT", configurator.options)
        self.assertEqual(configurator.options["DAGMAN_USE_STRICT"], 1)
        self.assertIsNone(configurator.config_path)
        self.assertIsNone(configurator.prefix)

    def testInitWrongOptionType(self):
        self.config[".wmsConfig.DAGMAN_USE_STRICT"] = "foo"
        with self.assertRaisesRegex(ValidationError, "DAGMAN_USE_STRICT".lower()):
            DagmanConfigurator(self.config)

    def testInitUnsupportedDagmanOption(self):
        """Test object instantiation with unsupported DAGMAN options."""
        self.config[".wmsConfig.DAGMAN_UNSUPPORTED_OPTION"] = "foo"
        with self.assertLogs(logger=logger, level="WARNING") as cm:
            configurator = DagmanConfigurator(self.config)
            self.assertIn("DAGMAN_UNSUPPORTED_OPTION", cm.output[0])
            self.assertNotIn("DAGMAN_UNSUPPORTED_OPTION", configurator.options)

    def testInitNoWmsConfig(self):
        """Test object instantiation fails when no WMS-specific options."""
        del self.config["wmsConfig"]
        with self.assertRaisesRegex(KeyError, "not found"):
            DagmanConfigurator(self.config)

    def testPrepare(self):
        """Test if the method creates the configuration file."""
        configurator = DagmanConfigurator(self.config)
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            configurator.prepare("dagman.conf", prefix=tmpdir)
            self.assertIn(Path(tmpdir), list(configurator.config_path.parents))
            self.assertTrue(configurator.config_path.is_file())
            self.assertEqual(configurator.config_path.read_text(), "DAGMAN_USE_STRICT = 0")

    def testPrepareWithUnsupportedOption(self):
        """Test if the method does not include unsupported options."""
        self.config[".wmsConfig.DAGMAN_UNSUPPORTED_OPTION"] = "foo"
        configurator = DagmanConfigurator(self.config)
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            configurator.prepare("dagman.conf", prefix=tmpdir)
            self.assertIn(Path(tmpdir), list(configurator.config_path.parents))
            self.assertTrue(configurator.config_path.is_file())
            self.assertEqual(configurator.config_path.read_text(), "DAGMAN_USE_STRICT = 0")

    def testPrepareConfigWriteFailure(self):
        """Test if the method raises when it can't create the configuration."""
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            os.chmod(tmpdir, 0o500)

            configurator = DagmanConfigurator(self.config)
            with self.assertLogs(logger=logger, level="ERROR") as cm, self.assertRaises(OSError):
                configurator.prepare("dagman.conf", f"{tmpdir}/subdir")
            self.assertIn("Could not write", cm.output[0])

            os.chmod(tmpdir, 0o700)

    def testConfigure(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            dag = HTCDag(name="test_configure")
            configurator = DagmanConfigurator(self.config)
            configurator.prepare("dagman.conf", prefix=tmpdir)
            configurator.configure(dag)
        self.assertIn("bps_wms_config_path", dag.graph["attr"])
        self.assertEqual(dag.graph["attr"]["bps_wms_config_path"], "dagman.conf")

    def testConfigureIfNotPrepared(self):
        """Test if the method raises when prepare step was skipped."""
        dag = HTCDag(name="test_configure_not_prepared")
        configurator = DagmanConfigurator(self.config)
        with self.assertRaisesRegex(RuntimeError, "file does not exist"):
            configurator.configure(dag)



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module enabling configuring DAGMan via submit YAML."""

__all__ = ["DagmanConfigurator"]

import logging
import os
from pathlib import Path
from typing import Any

import htcondor
from pydantic import AliasGenerator, ConfigDict, create_model

from lsst.ctrl.bps import BpsConfig

from .lssthtc import HTCDag

_LOG = logging.getLogger(__name__)

# Extract DAGMan configuration options with their types and default values from
# the local HTCondor configuration.
#
# Notes
# -----
# There are some DAGMan configuration options that names do not start with
# ``DAGMAN_`` (e.g., ``MAX_DAGMAN_LOG``). Hence, do not use
# ``key.startswith("DAGMAN_")``.
_fields = {key.lower(): (type(val), val) for key, val in htcondor.param.items() if "DAGMAN_" in key}

# Add some valid configuration options are not set by default by HTCondor and
# are missing from ``htcondor.param``.
#
# Notes
# -----
# A complete list of configuration options HTCondor supports can be found in
# ``src/condor_utils/param_info.in`` in
# `HTCondor GitHub repository <https://github.com/htcondor/htcondor>`_.
_fields.update(
    {
        "dagman_debug": (str, ""),
        "dagman_node_record_info": (str, ""),
        "dagman_record_machine_attrs": (str, ""),
    }
)

# Dynamically create a Pydantic model encapsulating the DAGMan configuration
# options gathered above.
_DagmanOptions = create_model(
    "DagmanOptions",
    __config__=ConfigDict(
        alias_generator=AliasGenerator(
            serialization_alias=lambda name: name.upper(),
        ),
        extra="allow",
        serialize_by_alias=True,
    ),
    **_fields,
)


class DagmanConfigurator:
    """Class responsible for setting WMS-specific configuration options.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration.
    search_opts : `dict` [`str`, `Any`], optional
        Options to use while searching the BPS configuration for values.

    Raises
    ------
    KeyError
        Raised if DAGMan configuration is missing from the BPS configuration.
    """

    def __init__(self, config: BpsConfig, search_opts: dict[str, Any] | None = None) -> None:
        if search_opts is None:
            search_opts = {}
        _, site = config.search("computeSite", search_opts)
        if site:
            search_opts["curvals"] = {"curr_site": site}
        _, wms_config = config.search("wmsConfig", search_opts)
        if not wms_config:
            raise KeyError("WMS-specific configuration not found")
        self._options = _DagmanOptions.model_validate({key.lower(): val for key, val in wms_config.items()})
        if self._options.model_extra:
            unknown_opts = [key.upper() for key in self._options.model_extra]
            _LOG.warning(
                "The following WMS-specific config options were not recognized and will be ignored: %s.",
                ", ".join(unknown_opts),
            )
        self.config_path: Path | None = None
        self.prefix: Path | None = None

    @property
    def options(self) -> dict[str, Any]:
        """DAGMan configuration options set via BPS (`dict` [`str`, `Any`])."""
        return {
            key: val
            for key, val in self._options.model_dump(exclude_unset=True).items()
            if key not in self._options.model_extra
        }

    def prepare(self, filename: os.PathLike | str, prefix: os.PathLike | str | None) -> None:
        """Write WMS-specific configuration to a file.

        Parameters
        ----------
        filename : `str`, optional
            Name of the file to use when creating the DAG configuration.
        prefix : `pathlib.Path` | `str`, optional
            Directory in which to output the DAG configuration file. If not
            provided, the script will be written to the current directory.

        Raises
        ------
        OSError
            Raised if the configuration file cannot be created.
        """
        if prefix:
            self.prefix = Path(prefix)
        self.config_path = self.prefix / filename if self.prefix else Path(filename)
        try:
            self.config_path.parent.mkdir(parents=True, exist_ok=True)
        except OSError as exc:
            _LOG.error(
                "Could not write WMS-specific configuration file '%s': %s",
                self.config_path,
                exc.strerror,
            )
            raise

        # Populate the DAG configuration file only with options that were
        # explicitly set in the BPS configuration.
        #
        # Notes
        # -----
        # The Pydantic model we are using to represent the DAGMan configuration
        # options allows for extra fields. However, it seems that
        # BaseModel.model_dump() does not support excluding these fields during
        # serialization at the moment (Pydantic ver. 2.12), so we have to do it
        # manually.
        self.config_path.write_text("\n".join(f"{key} = {val}" for key, val in self.options.items()))

    def configure(self, dag: HTCDag) -> None:
        """Add DAG configuration file to the workflow.

        Parameters
        ----------
        dag : `lsst.ctrl.bps.htcondor.HTCDag`
            HTCondor DAG.

        Raises
        ------
        RuntimeError
            Raised if the prepare step was omitted.

        Notes
        -----
        The path to the DAG configuration is added as a DAG attribute named
        ``bps_wms_config_path``. The stored path is relative to the prefix.
        """
        if self.config_path is None:
            raise RuntimeError(
                f"cannot add WMS-specific configuration to the workflow: file does not exist. "
                f"(hint: run {type(self).__qualname__}.prepare() to create it)"
            )
        config_path = self.config_path.relative_to(self.prefix) if self.prefix else self.config_path
        dag.add_attribs({"bps_wms_config_path": str(config_path)})



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for the report utilities."""

import logging
import os
import unittest
from pathlib import Path
from shutil import copy2, copytree

import htcondor

from lsst.ctrl.bps import (
    WmsSpecificInfo,
    WmsStates,
)
from lsst.ctrl.bps.htcondor import lssthtc, report_utils
from lsst.utils.tests import temporaryDirectory

logger = logging.getLogger("lsst.ctrl.bps.htcondor")

TESTDIR = os.path.abspath(os.path.dirname(__file__))

LOCATE_SUCCESS = """[
        CondorPlatform = "$CondorPlatform: X86_64-CentOS_7.9 $";
        MyType = "Scheduler";
        Machine = "testmachine";
        Name = "testmachine";
        CondorVersion = "$CondorVersion: 23.0.3 2024-04-04 $";
        MyAddress = "<127.0.0.1:9618?addrs=127.0.0.1-9618+snip>"
    ]
"""


class GetExitCodeSummaryTestCase(unittest.TestCase):
    """Test the function responsible for creating exit code summary."""

    def setUp(self):
        self.jobs = {
            "1.0": {
                "JobStatus": htcondor.JobStatus.IDLE,
                "bps_job_label": "foo",
            },
            "2.0": {
                "JobStatus": htcondor.JobStatus.RUNNING,
                "bps_job_label": "foo",
            },
            "3.0": {
                "JobStatus": htcondor.JobStatus.REMOVED,
                "bps_job_label": "foo",
            },
            "4.0": {
                "ExitCode": 0,
                "ExitBySignal": False,
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "bps_job_label": "bar",
            },
            "5.0": {
                "ExitCode": 1,
                "ExitBySignal": False,
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "bps_job_label": "bar",
            },
            "6.0": {
                "ExitBySignal": True,
                "ExitSignal": 11,
                "JobStatus": htcondor.JobStatus.HELD,
                "bps_job_label": "baz",
            },
            "7.0": {
                "ExitBySignal": False,
                "ExitCode": 42,
                "JobStatus": htcondor.JobStatus.HELD,
                "bps_job_label": "baz",
            },
            "8.0": {
                "JobStatus": htcondor.JobStatus.TRANSFERRING_OUTPUT,
                "bps_job_label": "qux",
            },
            "9.0": {
                "JobStatus": htcondor.JobStatus.SUSPENDED,
                "bps_job_label": "qux",
            },
        }

    def tearDown(self):
        pass

    def testMainScenario(self):
        actual = report_utils._get_exit_code_summary(self.jobs)
        expected = {"foo": [], "bar": [1], "baz": [11, 42], "qux": []}
        self.assertEqual(actual, expected)

    def testUnknownStatus(self):
        jobs = {
            "1.0": {
                "JobStatus": -1,
                "bps_job_label": "foo",
            }
        }
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            report_utils._get_exit_code_summary(jobs)
        self.assertIn("lsst.ctrl.bps.htcondor", cm.records[0].name)
        self.assertIn("Unknown", cm.output[0])
        self.assertIn("JobStatus", cm.output[0])

    def testUnknownKey(self):
        jobs = {
            "1.0": {
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "UnknownKey": None,
                "bps_job_label": "foo",
            }
        }
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            report_utils._get_exit_code_summary(jobs)
        self.assertIn("lsst.ctrl.bps.htcondor", cm.records[0].name)
        self.assertIn("Attribute", cm.output[0])
        self.assertIn("not found", cm.output[0])


class GetStateCountsFromDagJobTestCase(unittest.TestCase):
    """Test counting number of jobs per WMS state."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testCounts(self):
        job = {
            "DAG_NodesDone": 1,
            "DAG_JobsHeld": 2,
            "DAG_NodesFailed": 3,
            "DAG_NodesFutile": 4,
            "DAG_NodesQueued": 5,
            "DAG_NodesReady": 0,
            "DAG_NodesUnready": 7,
            "DAG_NodesTotal": 22,
        }

        truth = {
            WmsStates.SUCCEEDED: 1,
            WmsStates.HELD: 2,
            WmsStates.UNREADY: 7,
            WmsStates.READY: 0,
            WmsStates.FAILED: 3,
            WmsStates.PRUNED: 4,
            WmsStates.MISFIT: 0,
        }

        total, result = report_utils._get_state_counts_from_dag_job(job)
        self.assertEqual(total, 22)
        self.assertEqual(result, truth)


class GetInfoFromPathTestCase(unittest.TestCase):
    """Test _get_info_from_path function."""

    def test_tmpdir_abort(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_tmpdir_abort.dag.dagman.out", tmp_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, lssthtc.MISSING_ID)
            self.assertEqual(jobs, {})
            self.assertIn("Cannot submit from /tmp", message)

    def test_no_dagman_messages(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_no_messages.dag.dagman.out", tmp_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, lssthtc.MISSING_ID)
            self.assertEqual(jobs, {})
            self.assertIn("Could not find HTCondor files", message)

    def test_successful_run(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.log", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.out", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.nodes.log", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.node_status", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.info.json", tmp_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, "1163.0")
            self.assertEqual(len(jobs), 6)  # dag, pipetaskInit, 3 science, finalJob
            self.assertEqual(message, "")

    def test_relative_path(self):
        orig_dir = Path.cwd()
        with temporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            abs_path = Path(tmp_dir).resolve() / "subdir"
            abs_path.mkdir()
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.log", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.out", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.nodes.log", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.node_status", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.info.json", abs_path)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path("subdir")
            self.assertEqual(wms_workflow_id, "1163.0")
            self.assertEqual(len(jobs), 6)  # dag, pipetaskInit, 3 science, finalJob
            self.assertEqual(message, "")
            self.assertEqual(jobs["1163.0"]["Iwd"], str(abs_path))
            os.chdir(orig_dir)


class AddServiceJobSpecificInfoTestCase(unittest.TestCase):
    """Test _add_service_job_specific_info function.

    Note: The job_ad's are hardcoded in these tests.  The
    values in the dictionaries come from plugin code as
    well as HTCondor.  Changes in either of those codes
    that produce data for the job_ad can break this
    function without breaking these unit tests.

    Also, since hold status/messages stick around, testing
    various cases with and without job being held just to
    ensure get right status in both cases.
    """

    def testNotSubmitted(self):
        # Service job not submitted yet or can't be submitted.
        # (Typically an plugin bug.)
        # At this function level, can't tell if not submitted
        # yet or problem so it never will.
        job_ad = {
            "ClusterId": -64,
            "DAGManJobID": "8997.0",
            "DAGNodeName": "provisioningJob",
            "NodeStatus": lssthtc.NodeStatus.NOT_READY,
            "ProcId": 0,
            "bps_job_label": "service_provisioningJob",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "UNREADY", "status_details": ""}
        )

    def testRunning(self):
        # DAG hasn't completed (Running or held),
        # Service job is running.
        job_ad = {
            "ClusterId": 8523,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.RUNNING,
        }

        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "RUNNING", "status_details": ""}
        )

    def testDied(self):
        # DAG hasn't completed (Running or held),
        # Service job failed (completed non-zero exit code)
        job_ad = {
            "ClusterId": 8761,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.COMPLETED,
            "ExitCode": 4,
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "FAILED", "status_details": ""}
        )

    def testDeleted(self):
        # Deleted by user (never held)
        job_ad = {
            "ClusterId": 9086,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.REMOVED,
            "ProcId": 0,
            "Reason": "via condor_rm (by user mgower)",
            "job_evicted_time": "2025-02-11T11:35:04",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "DELETED", "status_details": ""}
        )

    def testSucceedEarly(self):
        # DAG hasn't completed (Running or held),
        # Service job completed with exit code 0
        job_ad = {
            "ClusterId": 8761,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.COMPLETED,
            "ExitCode": 0,
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context,
            {
                "job_name": "provisioningJob",
                "status": "SUCCEEDED",
                "status_details": "(Note: Finished before workflow.)",
            },
        )

    def testSucceedOldRemoveMessage(self):
        # DAG completed, job was in running state when removed.
        job_ad = {
            "ClusterId": 8761,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.REMOVED,
            "Reason": "Removed by DAGMan (by user mgower)",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "SUCCEEDED", "status_details": ""}
        )

    def testSucceed(self):
        # DAG completed, job was in running state when removed.
        job_ad = {
            "ClusterId": 8761,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.REMOVED,
            "Reason": (
                "removed because <OtherJobRemoveRequirements = DAGManJobId =?= 8556>"
                " fired when job (8556.0) was removed"
            ),
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "SUCCEEDED", "status_details": ""}
        )

    def testUserHeldWhileRunning(self):
        # DAG hasn't completed (Running or held),
        # user put at least service job on hold
        job_ad = {
            "ClusterId": 8523,
            "ProcId": 0,
            "DAGNodeName": "provisioningJob",
            "JobStatus": htcondor.JobStatus.HELD,
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
        }

        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context,
            {
                "job_name": "provisioningJob",
                "status": "HELD",
                "status_details": "(via condor_hold (by user mgower))",
            },
        )

    def testHeldByHTC(self):
        # Job put on hold by HTCondor, removed when DAG ends
        job_ad = {
            "ClusterId": 8693,
            "DAGNodeName": "provisioningJob",
            "HoldReason": "Failed to execute",
            "HoldReasonCode": 6,
            "HoldReasonSubCode": 2,
            "JobStatus": htcondor.JobStatus.REMOVED,
            "ProcId": 0,
            "Reason": "Removed by DAGMan (by user mgower)",
            "job_held_time": "2025-02-07T12:50:07",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context,
            {
                "job_name": "provisioningJob",
                "status": "DELETED",
                "status_details": "(Job was held for the following reason: Failed to execute)",
            },
        )

    def testHeldReleasedRunning(self):
        # DAG hasn't completed (Running or held),
        # Since held info will be in job_ad, make sure knows released.
        job_ad = {
            "ClusterId": 8625,
            "DAGNodeName": "provisioningJob",
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.RUNNING,
            "LogNotes": "DAG Node: provisioningJob",
            "ProcId": 0,
            "job_held_time": "2025-02-07T12:33:34",
            "job_released_time": "2025-02-07T12:33:47",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "RUNNING", "status_details": ""}
        )

    def testHeldReleasedDied(self):
        # Since held info will be in job_ad,
        # make sure knows status after released.
        job_ad = {
            "ClusterId": 9120,
            "DAGNodeName": "provisioningJob",
            "ExitBySignal": False,
            "ExitCode": 4,
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.COMPLETED,
            "ProcId": 0,
            "Reason": "via condor_release (by user mgower)",
            "ReturnValue": 4,
            "TerminatedNormally": True,
            "job_held_time": "2025-02-11T11:46:40",
            "job_released_time": "2025-02-11T11:46:47",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "FAILED", "status_details": ""}
        )

    def testHeldReleasedSuccessEarly(self):
        # Since held info will be in job_ad,
        # make sure knows status after released.
        job_ad = {
            "ClusterId": 9154,
            "DAGNodeName": "provisioningJob",
            "ExitBySignal": False,
            "ExitCode": 0,
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.COMPLETED,
            "ProcId": 0,
            "Reason": "via condor_release (by user mgower)",
            "TerminatedNormally": True,
            "job_held_time": "2025-02-11T11:55:20",
            "job_released_time": "2025-02-11T11:55:25",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context,
            {
                "job_name": "provisioningJob",
                "status": "SUCCEEDED",
                "status_details": "(Note: Finished before workflow.)",
            },
        )

    def testHeldReleasedSuccess(self):
        # DAG has completed.
        # Since held info will be in job_ad,
        # make sure knows status after released.
        job_ad = {
            "ClusterId": 8625,
            "DAGNodeName": "provisioningJob",
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.REMOVED,
            "ProcId": 0,
            "Reason": "removed because <OtherJobRemoveRequirements = DAGManJobId =?= "
            "8624> fired when job (8624.0) was removed",
            "job_held_time": "2025-02-07T12:33:34",
            "job_released_time": "2025-02-07T12:33:47",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "SUCCEEDED", "status_details": ""}
        )

    def testHeldReleasedDeleted(self):
        # Since held info will be in job_ad,
        # make sure knows status after released.
        job_ad = {
            "ClusterId": 9086,
            "DAGNodeName": "provisioningJob",
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.REMOVED,
            "ProcId": 0,
            "Reason": "via condor_rm (by user mgower)",
            "job_evicted_time": "2025-02-11T11:35:04",
            "job_held_time": "2025-02-11T11:35:04",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context, {"job_name": "provisioningJob", "status": "DELETED", "status_details": ""}
        )

    def testHeldReleasedHeld(self):
        # Since release info will be in job_ad,
        # make sure knows held after release.
        job_ad = {
            "ClusterId": 8659,
            "DAGNodeName": "provisioningJob",
            "HoldReason": "via condor_hold (by user mgower)",
            "HoldReasonCode": 1,
            "HoldReasonSubCode": 0,
            "JobStatus": htcondor.JobStatus.REMOVED,
            "ProcId": 0,
            "Reason": "Removed by DAGMan (by user mgower)",
            "TerminatedNormally": False,
            "job_held_time": "2025-02-07T12:36:15",
            "job_released_time": "2025-02-07T12:36:07",
        }
        results = WmsSpecificInfo()
        report_utils._add_service_job_specific_info(job_ad, results)
        self.assertEqual(
            results.context,
            {
                "job_name": "provisioningJob",
                "status": "DELETED",
                "status_details": "(Job was held for the following reason: via condor_hold (by user mgower))",
            },
        )


class GetRunSummaryTestCase(unittest.TestCase):
    """Test _get_run_summary function."""

    def testJobSummaryInJobAd(self):
        summary = "pipetaskInit:1;label1:2;label2:2;finalJob:1"
        job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "bps_job_summary": summary}
        results = report_utils._get_run_summary(job_ad)
        self.assertEqual(results, summary)

    def testRunSummaryInJobAd(self):
        summary = "pipetaskInit:1;label1:2;label2:2;finalJob:1"
        job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "bps_run_summary": summary}
        results = report_utils._get_run_summary(job_ad)
        self.assertEqual(results, summary)

    def testSummaryFromDag(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/good.dag", tmp_dir)
            job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "Iwd": tmp_dir}
            results = report_utils._get_run_summary(job_ad)
            self.assertEqual(results, "pipetaskInit:1;label1:1;label2:1;label3:1;finalJob:1")

    def testSummaryNoDag(self):
        with self.assertLogs(logger=logger, level="WARNING") as cm:
            with temporaryDirectory() as tmp_dir:
                job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "Iwd": tmp_dir}
                results = report_utils._get_run_summary(job_ad)
                self.assertEqual(results, "")
            self.assertIn("lsst.ctrl.bps.htcondor", cm.records[0].name)
            self.assertIn("Could not get run summary for htcondor job", cm.output[0])


class IsServiceJobTestCase(unittest.TestCase):
    """Test is_service_job function."""

    def testNotServiceJob(self):
        job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "wms_node_type": lssthtc.WmsNodeType.PAYLOAD}
        self.assertFalse(report_utils.is_service_job(job_ad))

    def testIsServiceJob(self):
        job_ad = {"ClusterId": 8659, "DAGNodeName": "testJob", "wms_node_type": lssthtc.WmsNodeType.SERVICE}
        self.assertTrue(report_utils.is_service_job(job_ad))

    def testMissingBpsType(self):
        job_ad = {
            "ClusterId": 8659,
            "DAGNodeName": "testJob",
        }
        self.assertFalse(report_utils.is_service_job(job_ad))


class CreateDetailedReportFromJobsTestCase(unittest.TestCase):
    """Test _create_detailed_report_from_jobs function."""

    def testTinySuccess(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "tiny_success")
            copytree(f"{TESTDIR}/data/tiny_success", test_submit_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(test_submit_dir)
            run_reports = report_utils._create_detailed_report_from_jobs(wms_workflow_id, jobs)
            self.assertEqual(len(run_reports), 1)
            report = run_reports[wms_workflow_id]
            self.assertEqual(report.wms_id, wms_workflow_id)
            self.assertEqual(report.state, WmsStates.SUCCEEDED)
            self.assertTrue(os.path.samefile(report.path, test_submit_dir))
            self.assertEqual(report.run_summary, "pipetaskInit:1;label1:1;label2:1;finalJob:1")
            self.assertEqual(
                report.job_state_counts,
                {
                    WmsStates.UNKNOWN: 0,
                    WmsStates.MISFIT: 0,
                    WmsStates.UNREADY: 0,
                    WmsStates.READY: 0,
                    WmsStates.PENDING: 0,
                    WmsStates.RUNNING: 0,
                    WmsStates.DELETED: 0,
                    WmsStates.HELD: 0,
                    WmsStates.SUCCEEDED: 4,
                    WmsStates.FAILED: 0,
                    WmsStates.PRUNED: 0,
                },
            )
            self.assertEqual(
                report.specific_info.context,
                {"job_name": "provisioningJob", "status": "SUCCEEDED", "status_details": ""},
            )

    def testTinyProblems(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "tiny_problems")
            copytree(f"{TESTDIR}/data/tiny_problems", test_submit_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(test_submit_dir)
            run_reports = report_utils._create_detailed_report_from_jobs(wms_workflow_id, jobs)
            self.assertEqual(len(run_reports), 1)
            report = run_reports[wms_workflow_id]
            self.assertEqual(report.wms_id, wms_workflow_id)
            self.assertEqual(report.state, WmsStates.FAILED)
            self.assertTrue(os.path.samefile(report.path, test_submit_dir))
            self.assertEqual(report.run_summary, "pipetaskInit:1;label1:2;label2:2;finalJob:1")
            self.assertEqual(
                report.job_state_counts,
                {
                    WmsStates.UNKNOWN: 0,
                    WmsStates.MISFIT: 0,
                    WmsStates.UNREADY: 0,
                    WmsStates.READY: 0,
                    WmsStates.PENDING: 0,
                    WmsStates.RUNNING: 0,
                    WmsStates.DELETED: 0,
                    WmsStates.HELD: 0,
                    WmsStates.SUCCEEDED: 4,
                    WmsStates.FAILED: 1,
                    WmsStates.PRUNED: 1,
                },
            )
            self.assertEqual(
                run_reports[wms_workflow_id].specific_info.context,
                {"job_name": "provisioningJob", "status": "SUCCEEDED", "status_details": ""},
            )

    def testTinyRunning(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "tiny_running")
            copytree(f"{TESTDIR}/data/tiny_running", test_submit_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(test_submit_dir)
            run_reports = report_utils._create_detailed_report_from_jobs(wms_workflow_id, jobs)
            self.assertEqual(len(run_reports), 1)
            report = run_reports[wms_workflow_id]
            self.assertEqual(report.wms_id, wms_workflow_id)
            self.assertEqual(report.state, WmsStates.RUNNING)
            self.assertTrue(os.path.samefile(report.path, test_submit_dir))
            self.assertEqual(report.run_summary, "pipetaskInit:1;label1:1;label2:1;finalJob:1")
            self.assertEqual(
                report.job_state_counts,
                {
                    WmsStates.UNKNOWN: 0,
                    WmsStates.MISFIT: 0,
                    WmsStates.UNREADY: 2,
                    WmsStates.READY: 0,
                    WmsStates.PENDING: 0,
                    WmsStates.RUNNING: 1,
                    WmsStates.DELETED: 0,
                    WmsStates.HELD: 0,
                    WmsStates.SUCCEEDED: 1,
                    WmsStates.FAILED: 0,
                    WmsStates.PRUNED: 0,
                },
            )
            self.assertEqual(
                report.specific_info.context,
                {"job_name": "provisioningJob", "status": "RUNNING", "status_details": ""},
            )

    def testNoopRunning(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "noop_running_1")
            copytree(f"{TESTDIR}/data/noop_running_1", test_submit_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(test_submit_dir)
            run_reports = report_utils._create_detailed_report_from_jobs(wms_workflow_id, jobs)
            self.assertEqual(len(run_reports), 1)
            report = run_reports[wms_workflow_id]
            self.assertEqual(report.wms_id, wms_workflow_id)
            self.assertEqual(report.state, WmsStates.RUNNING)
            self.assertTrue(os.path.samefile(report.path, test_submit_dir))
            self.assertEqual(
                set(report.run_summary.split(";")),
                {"pipetaskInit:1", "label1:6", "label2:6", "label3:6", "label4:6", "label5:6", "finalJob:1"},
            )
            self.assertEqual(
                report.job_state_counts,
                {
                    WmsStates.UNKNOWN: 0,
                    WmsStates.MISFIT: 0,
                    WmsStates.UNREADY: 12,
                    WmsStates.READY: 0,
                    WmsStates.PENDING: 1,
                    WmsStates.RUNNING: 10,
                    WmsStates.DELETED: 0,
                    WmsStates.HELD: 0,
                    WmsStates.SUCCEEDED: 9,
                    WmsStates.FAILED: 0,
                    WmsStates.PRUNED: 0,
                },
            )
            self.assertEqual(report.total_number_jobs, 32)
            self.assertIsNone(report.specific_info)

    def testNoopFailed(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "noop_failed_1")
            copytree(f"{TESTDIR}/data/noop_failed_1", test_submit_dir)
            wms_workflow_id, jobs, message = report_utils._get_info_from_path(test_submit_dir)
            run_reports = report_utils._create_detailed_report_from_jobs(wms_workflow_id, jobs)
            self.assertEqual(len(run_reports), 1)
            report = run_reports[wms_workflow_id]
            self.assertEqual(report.wms_id, wms_workflow_id)
            self.assertEqual(report.state, WmsStates.FAILED)
            self.assertTrue(os.path.samefile(report.path, test_submit_dir))
            self.assertEqual(
                set(report.run_summary.split(";")),
                {"pipetaskInit:1", "label1:6", "label2:6", "label3:6", "label4:6", "label5:6", "finalJob:1"},
            )
            self.assertEqual(
                report.job_state_counts,
                {
                    WmsStates.UNKNOWN: 0,
                    WmsStates.MISFIT: 0,
                    WmsStates.UNREADY: 0,
                    WmsStates.READY: 0,
                    WmsStates.PENDING: 0,
                    WmsStates.RUNNING: 0,
                    WmsStates.DELETED: 0,
                    WmsStates.HELD: 0,
                    WmsStates.SUCCEEDED: 27,
                    WmsStates.FAILED: 1,
                    WmsStates.PRUNED: 4,
                },
            )
            self.assertEqual(report.total_number_jobs, 32)
            self.assertIsNone(report.specific_info)
            self.assertEqual(
                report.exit_code_summary,
                {
                    "pipetaskInit": [],
                    "label1": [],
                    "label2": [1],
                    "label3": [],
                    "label4": [],
                    "label5": [],
                    "finalJob": [],
                },
            )


class GetStatusFromIdTestCase(unittest.TestCase):
    """Test _get_status_from_id function."""

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.report_utils._get_info_from_schedd")
    def testNotFound(self, mock_get):
        mock_get.return_value = {}

        state, message = report_utils._get_status_from_id("100", 0, {})

        mock_get.assert_called_once_with("100", 0, {})

        self.assertEqual(state, WmsStates.UNKNOWN)
        self.assertEqual(message, "DAGMan job 100 not found in queue or history.  Check id or try path.")

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.report_utils._htc_status_to_wms_state")
    @unittest.mock.patch("lsst.ctrl.bps.htcondor.report_utils._get_info_from_schedd")
    def testFound(self, mock_get, mock_conversion):
        fake_id = "100.0"
        dag_ads = {fake_id: {"JobStatus": lssthtc.JobStatus.RUNNING}}
        mock_get.return_value = {"schedd1": dag_ads}
        mock_conversion.return_value = WmsStates.RUNNING

        state, message = report_utils._get_status_from_id(fake_id, 0, {})

        mock_get.assert_called_once_with(fake_id, 0, {})
        mock_conversion.assert_called_once_with(dag_ads[fake_id])

        self.assertEqual(state, WmsStates.RUNNING)
        self.assertEqual(message, "")


class GetStatusFromPathTestCase(unittest.TestCase):
    """Test _get_status_from_path function."""

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.report_utils.read_dag_log")
    def testNoDagLog(self, mock_read):
        mock_read.side_effect = FileNotFoundError

        fake_path = "/fake/path"
        state, message = report_utils._get_status_from_path(fake_path)

        mock_read.assert_called_once_with(Path(fake_path))

        self.assertEqual(state, WmsStates.UNKNOWN)
        self.assertEqual(message, f"DAGMan log not found in {fake_path}.  Check path.")

    def testSuccess(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "tiny_success")
            copytree(f"{TESTDIR}/data/tiny_success", test_submit_dir)
            state, message = report_utils._get_status_from_path(test_submit_dir)

            self.assertEqual(state, WmsStates.SUCCEEDED)
            self.assertEqual(message, "")

    def testFailure(self):
        with temporaryDirectory() as tmp_dir:
            test_submit_dir = os.path.join(tmp_dir, "tiny_problems")
            copytree(f"{TESTDIR}/data/tiny_problems", test_submit_dir)
            state, message = report_utils._get_status_from_path(test_submit_dir)

            self.assertEqual(state, WmsStates.FAILED)
            self.assertEqual(message, "")


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Interface between generic workflow to HTCondor workflow system."""

__all__ = ["HTCondorService"]


import logging
import os
from pathlib import Path

import htcondor
from packaging import version

from lsst.ctrl.bps import (
    BaseWmsService,
    WmsStates,
)
from lsst.ctrl.bps.bps_utils import chdir
from lsst.daf.butler import Config
from lsst.utils.timer import time_this

from .common_utils import WmsIdType, _wms_id_to_cluster, _wms_id_to_dir, _wms_id_type
from .dagman_configurator import DagmanConfigurator
from .htcondor_config import HTC_DEFAULTS_URI
from .htcondor_workflow import HTCondorWorkflow
from .lssthtc import (
    _locate_schedds,
    _update_rescue_file,
    condor_q,
    htc_backup_files,
    htc_create_submit_from_cmd,
    htc_create_submit_from_dag,
    htc_create_submit_from_file,
    htc_submit_dag,
    htc_version,
    read_dag_status,
    write_dag_info,
)
from .provisioner import Provisioner
from .report_utils import (
    _get_status_from_id,
    _get_status_from_path,
    _report_from_id,
    _report_from_path,
    _summary_report,
)

_LOG = logging.getLogger(__name__)


class HTCondorService(BaseWmsService):
    """HTCondor version of WMS service."""

    @property
    def defaults(self):
        return Config(HTC_DEFAULTS_URI)

    @property
    def defaults_uri(self):
        return HTC_DEFAULTS_URI

    def prepare(self, config, generic_workflow, out_prefix=None):
        """Convert generic workflow to an HTCondor DAG ready for submission.

        Parameters
        ----------
        config : `lsst.ctrl.bps.BpsConfig`
            BPS configuration that includes necessary submit/runtime
            information.
        generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
            The generic workflow (e.g., has executable name and arguments).
        out_prefix : `str`
            The root directory into which all WMS-specific files are written.

        Returns
        -------
        workflow : `lsst.ctrl.bps.wms.htcondor.HTCondorWorkflow`
            HTCondor workflow ready to be run.
        """
        _LOG.debug("out_prefix = '%s'", out_prefix)
        with time_this(log=_LOG, level=logging.INFO, prefix=None, msg="Completed HTCondor workflow creation"):
            workflow = HTCondorWorkflow.from_generic_workflow(
                config,
                generic_workflow,
                out_prefix,
                f"{self.__class__.__module__}.{self.__class__.__name__}",
            )

            _, enable_provisioning = config.search("provisionResources")
            if enable_provisioning:
                provisioner = Provisioner(config)
                provisioner.configure()
                provisioner.prepare("provisioningJob.bash", prefix=out_prefix)
                provisioner.provision(workflow.dag)

            try:
                configurator = DagmanConfigurator(config)
            except KeyError:
                _LOG.debug(
                    "No DAGMan-specific settings were found in BPS config; "
                    "skipping writing DAG-specific configuration file."
                )
            else:
                configurator.prepare("dagman.conf", prefix=out_prefix)
                configurator.configure(workflow.dag)

        with time_this(
            log=_LOG, level=logging.INFO, prefix=None, msg="Completed writing out HTCondor workflow"
        ):
            workflow.write(out_prefix)
        return workflow

    def submit(self, workflow, **kwargs):
        """Submit a single HTCondor workflow.

        Parameters
        ----------
        workflow : `lsst.ctrl.bps.BaseWorkflow`
            A single HTCondor workflow to submit.  run_id is updated after
            successful submission to WMS.
        **kwargs : `~typing.Any`
            Keyword arguments for the options.
        """
        dag = workflow.dag
        ver = version.parse(htc_version())

        # For workflow portability, internal paths are all relative. Hence
        # the DAG needs to be submitted to HTCondor from inside the submit
        # directory.
        with chdir(workflow.submit_path):
            try:
                if ver >= version.parse("8.9.3"):
                    sub = htc_create_submit_from_dag(dag.graph["dag_filename"], dag.graph["submit_options"])
                else:
                    sub = htc_create_submit_from_cmd(dag.graph["dag_filename"], dag.graph["submit_options"])
            except Exception:
                _LOG.error(
                    "Problems creating HTCondor submit object from filename: %s", dag.graph["dag_filename"]
                )
                raise

            _LOG.info("Submitting from directory: %s", os.getcwd())
            schedd_dag_info = htc_submit_dag(sub)
            if schedd_dag_info:
                write_dag_info(f"{dag.name}.info.json", schedd_dag_info)

                _, dag_info = schedd_dag_info.popitem()
                _, dag_ad = dag_info.popitem()

                dag.run_id = f"{dag_ad['ClusterId']}.{dag_ad['ProcId']}"
                workflow.run_id = dag.run_id
            else:
                raise RuntimeError("Submission failed: unable to retrieve DAGMan job information")

    def restart(self, wms_workflow_id):
        """Restart a failed DAGMan workflow.

        Parameters
        ----------
        wms_workflow_id : `str`
            The directory with HTCondor files.

        Returns
        -------
        run_id : `str`
            HTCondor id of the restarted DAGMan job. If restart failed, it will
            be set to None.
        run_name : `str`
            Name of the restarted workflow. If restart failed, it will be set
            to None.
        message : `str`
            A message describing any issues encountered during the restart.
            If there were no issues, an empty string is returned.
        """
        wms_path, id_type = _wms_id_to_dir(wms_workflow_id)
        if wms_path is None:
            return (
                None,
                None,
                (
                    f"workflow with run id '{wms_workflow_id}' not found. "
                    "Hint: use run's submit directory as the id instead"
                ),
            )

        if id_type in {WmsIdType.GLOBAL, WmsIdType.LOCAL}:
            if not wms_path.is_dir():
                return None, None, f"submit directory '{wms_path}' for run id '{wms_workflow_id}' not found."

        _LOG.info("Restarting workflow from directory '%s'", wms_path)
        rescue_dags = list(wms_path.glob("*.dag.rescue*"))
        if not rescue_dags:
            return None, None, f"HTCondor rescue DAG(s) not found in '{wms_path}'"

        _LOG.info("Verifying that the workflow is not already in the job queue")
        schedd_dag_info = condor_q(constraint=f'regexp("dagman$", Cmd) && Iwd == "{wms_path}"')
        if schedd_dag_info:
            _, dag_info = schedd_dag_info.popitem()
            _, dag_ad = dag_info.popitem()
            id_ = dag_ad["GlobalJobId"]
            return None, None, f"Workflow already in the job queue (global job id: '{id_}')"

        _LOG.info("Checking execution status of the workflow")
        warn = False
        dag_ad = read_dag_status(str(wms_path))
        if dag_ad:
            nodes_total = dag_ad.get("NodesTotal", 0)
            if nodes_total != 0:
                nodes_done = dag_ad.get("NodesDone", 0)
                if nodes_total == nodes_done:
                    return None, None, "All jobs in the workflow finished successfully"
            else:
                warn = True
        else:
            warn = True
        if warn:
            _LOG.warning(
                "Cannot determine the execution status of the workflow, continuing with restart regardless"
            )

        _LOG.info("Backing up select HTCondor files from previous run attempt")
        rescue_file = htc_backup_files(wms_path, subdir="backups")
        if (wms_path / "subdags").exists():
            _update_rescue_file(rescue_file)

        # For workflow portability, internal paths are all relative. Hence
        # the DAG needs to be resubmitted to HTCondor from inside the submit
        # directory.
        _LOG.info("Adding workflow to the job queue")
        run_id, run_name, message = None, None, ""
        with chdir(wms_path):
            try:
                dag_path = next(Path.cwd().glob("*.dag.condor.sub"))
            except StopIteration:
                message = f"DAGMan submit description file not found in '{wms_path}'"
            else:
                sub = htc_create_submit_from_file(dag_path.name)
                schedd_dag_info = htc_submit_dag(sub)

                # Save select information about the DAGMan job to a file. Use
                # the run name (available in the ClassAd) as the filename.
                if schedd_dag_info:
                    dag_info = next(iter(schedd_dag_info.values()))
                    dag_ad = next(iter(dag_info.values()))
                    write_dag_info(f"{dag_ad['bps_run']}.info.json", schedd_dag_info)
                    run_id = f"{dag_ad['ClusterId']}.{dag_ad['ProcId']}"
                    run_name = dag_ad["bps_run"]
                else:
                    message = "DAGMan job information unavailable"

        return run_id, run_name, message

    def list_submitted_jobs(self, wms_id=None, user=None, require_bps=True, pass_thru=None, is_global=False):
        """Query WMS for list of submitted WMS workflows/jobs.

        This should be a quick lookup function to create list of jobs for
        other functions.

        Parameters
        ----------
        wms_id : `int` or `str`, optional
            Id or path that can be used by WMS service to look up job.
        user : `str`, optional
            User whose submitted jobs should be listed.
        require_bps : `bool`, optional
            Whether to require jobs returned in list to be bps-submitted jobs.
        pass_thru : `str`, optional
            Information to pass through to WMS.
        is_global : `bool`, optional
            If set, all job queues (and their histories) will be queried for
            job information. Defaults to False which means that only the local
            job queue will be queried.

        Returns
        -------
        job_ids : `list` [`~typing.Any`]
            Only job ids to be used by cancel and other functions.  Typically
            this means top-level jobs (i.e., not children jobs).
        """
        _LOG.debug(
            "list_submitted_jobs params: wms_id=%s, user=%s, require_bps=%s, pass_thru=%s, is_global=%s",
            wms_id,
            user,
            require_bps,
            pass_thru,
            is_global,
        )

        # Determine which Schedds will be queried for job information.
        coll = htcondor.Collector()

        schedd_ads = []
        if is_global:
            schedd_ads.extend(coll.locateAll(htcondor.DaemonTypes.Schedd))
        else:
            schedd_ads.append(coll.locate(htcondor.DaemonTypes.Schedd))

        # Construct appropriate constraint expression using provided arguments.
        constraint = "False"
        if wms_id is None:
            if user is not None:
                constraint = f'(Owner == "{user}")'
        else:
            schedd_ad, cluster_id, id_type = _wms_id_to_cluster(wms_id)
            if cluster_id is not None:
                constraint = f"(DAGManJobId == {cluster_id} || ClusterId == {cluster_id})"

                # If provided id is either a submission path or a global id,
                # make sure the right Schedd will be queried regardless of
                # 'is_global' value.
                if id_type in {WmsIdType.GLOBAL, WmsIdType.PATH}:
                    schedd_ads = [schedd_ad]
        if require_bps:
            constraint += ' && (bps_isjob == "True")'
        if pass_thru:
            if "-forcex" in pass_thru:
                pass_thru_2 = pass_thru.replace("-forcex", "")
                if pass_thru_2 and not pass_thru_2.isspace():
                    constraint += f" && ({pass_thru_2})"
            else:
                constraint += f" && ({pass_thru})"

        # Create a list of scheduler daemons which need to be queried.
        schedds = {ad["Name"]: htcondor.Schedd(ad) for ad in schedd_ads}

        _LOG.debug("constraint = %s, schedds = %s", constraint, ", ".join(schedds))
        results = condor_q(constraint=constraint, schedds=schedds)

        # Prune child jobs where DAG job is in queue (i.e., aren't orphans).
        job_ids = []
        for job_info in results.values():
            for job_id, job_ad in job_info.items():
                _LOG.debug("job_id=%s DAGManJobId=%s", job_id, job_ad.get("DAGManJobId", "None"))
                if "DAGManJobId" not in job_ad:
                    job_ids.append(job_ad.get("GlobalJobId", job_id))
                else:
                    _LOG.debug("Looking for %s", f"{job_ad['DAGManJobId']}.0")
                    _LOG.debug("\tin jobs.keys() = %s", job_info.keys())
                    if f"{job_ad['DAGManJobId']}.0" not in job_info:  # orphaned job
                        job_ids.append(job_ad.get("GlobalJobId", job_id))

        _LOG.debug("job_ids = %s", job_ids)
        return job_ids

    def get_status(
        self,
        wms_workflow_id: str,
        hist: float = 1,
        is_global: bool = False,
    ) -> tuple[WmsStates, str]:
        """Return status of run based upon given constraints.

        Parameters
        ----------
        wms_workflow_id : `str`
            Limit to specific run based on id (queue id or path).
        hist : `float`, optional
            Limit history search to this many days. Defaults to 1.
        is_global : `bool`, optional
            If set, all job queues (and their histories) will be queried for
            job information. Defaults to False which means that only the local
            job queue will be queried.

        Returns
        -------
        state : `lsst.ctrl.bps.WmsStates`
            Status of single run from given information.
        message : `str`
            Extra message for status command to print.  This could be pointers
            to documentation or to WMS specific commands.
        """
        _LOG.debug("get_status: id=%s, hist=%s, is_global=%s", wms_workflow_id, hist, is_global)

        id_type = _wms_id_type(wms_workflow_id)
        _LOG.debug("id_type = %s", id_type.name)

        if id_type == WmsIdType.LOCAL:
            schedulers = _locate_schedds(locate_all=is_global)
            _LOG.debug("schedulers = %s", schedulers)
            state, message = _get_status_from_id(wms_workflow_id, hist, schedds=schedulers)
        elif id_type == WmsIdType.GLOBAL:
            schedulers = _locate_schedds(locate_all=True)
            _LOG.debug("schedulers = %s", schedulers)
            state, message = _get_status_from_id(wms_workflow_id, hist, schedds=schedulers)
        elif id_type == WmsIdType.PATH:
            state, message = _get_status_from_path(wms_workflow_id)
        else:
            state, message = WmsStates.UNKNOWN, "Invalid job id"
        _LOG.debug("state: %s, %s", state, message)

        return state, message

    def report(
        self,
        wms_workflow_id=None,
        user=None,
        hist=0,
        pass_thru=None,
        is_global=False,
        return_exit_codes=False,
    ):
        """Return run information based upon given constraints.

        Parameters
        ----------
        wms_workflow_id : `str`, optional
            Limit to specific run based on id.
        user : `str`, optional
            Limit results to runs for this user.
        hist : `float`, optional
            Limit history search to this many days. Defaults to 0.
        pass_thru : `str`, optional
            Constraints to pass through to HTCondor.
        is_global : `bool`, optional
            If set, all job queues (and their histories) will be queried for
            job information. Defaults to False which means that only the local
            job queue will be queried.
        return_exit_codes : `bool`, optional
            If set, return exit codes related to jobs with a
            non-success status. Defaults to False, which means that only
            the summary state is returned.

            Only applicable in the context of a WMS with associated
            handlers to return exit codes from jobs.

        Returns
        -------
        runs : `list` [`lsst.ctrl.bps.WmsRunReport`]
            Information about runs from given job information.
        message : `str`
            Extra message for report command to print.  This could be pointers
            to documentation or to WMS specific commands.
        """
        if wms_workflow_id:
            id_type = _wms_id_type(wms_workflow_id)
            if id_type == WmsIdType.LOCAL:
                schedulers = _locate_schedds(locate_all=is_global)
                run_reports, message = _report_from_id(wms_workflow_id, hist, schedds=schedulers)
            elif id_type == WmsIdType.GLOBAL:
                schedulers = _locate_schedds(locate_all=True)
                run_reports, message = _report_from_id(wms_workflow_id, hist, schedds=schedulers)
            elif id_type == WmsIdType.PATH:
                run_reports, message = _report_from_path(wms_workflow_id)
            else:
                run_reports, message = {}, "Invalid job id"
        else:
            schedulers = _locate_schedds(locate_all=is_global)
            run_reports, message = _summary_report(user, hist, pass_thru, schedds=schedulers)
        _LOG.debug("report: %s, %s", run_reports, message)

        return list(run_reports.values()), message

    def cancel(self, wms_id, pass_thru=None):
        """Cancel submitted workflows/jobs.

        Parameters
        ----------
        wms_id : `str`
            Id or path of job that should be canceled.
        pass_thru : `str`, optional
            Information to pass through to WMS.

        Returns
        -------
        deleted : `bool`
            Whether successful deletion or not.  Currently, if any doubt or any
            individual jobs not deleted, return False.
        message : `str`
            Any message from WMS (e.g., error details).
        """
        _LOG.debug("Canceling wms_id = %s", wms_id)

        schedd_ad, cluster_id, _ = _wms_id_to_cluster(wms_id)

        if cluster_id is None:
            deleted = False
            message = "invalid id"
        else:
            _LOG.debug(
                "Canceling job managed by schedd_name = %s with cluster_id = %s",
                cluster_id,
                schedd_ad["Name"],
            )
            schedd = htcondor.Schedd(schedd_ad)

            constraint = f"ClusterId == {cluster_id}"
            if pass_thru is not None and "-forcex" in pass_thru:
                pass_thru_2 = pass_thru.replace("-forcex", "")
                if pass_thru_2 and not pass_thru_2.isspace():
                    constraint += f"&& ({pass_thru_2})"
                _LOG.debug("JobAction.RemoveX constraint = %s", constraint)
                results = schedd.act(htcondor.JobAction.RemoveX, constraint)
            else:
                if pass_thru:
                    constraint += f"&& ({pass_thru})"
                _LOG.debug("JobAction.Remove constraint = %s", constraint)
                results = schedd.act(htcondor.JobAction.Remove, constraint)
            _LOG.debug("Remove results: %s", results)

            if results["TotalSuccess"] > 0 and results["TotalError"] == 0:
                deleted = True
                message = ""
            else:
                deleted = False
                if results["TotalSuccess"] == 0 and results["TotalError"] == 0:
                    message = "no such bps job in batch queue"
                else:
                    message = f"unknown problems deleting: {results}"

        _LOG.debug("deleted: %s; message = %s", deleted, message)
        return deleted, message

    def ping(self, pass_thru):
        """Check whether WMS services are up, reachable, and can authenticate
        if authentication is required.

        The services to be checked are those needed for submit, report, cancel,
        restart, but ping cannot guarantee whether jobs would actually run
        successfully.

        Parameters
        ----------
        pass_thru : `str`, optional
            Information to pass through to WMS.

        Returns
        -------
        status : `int`
            0 for success, non-zero for failure.
        message : `str`
            Any message from WMS (e.g., error details).
        """
        coll = htcondor.Collector()
        secman = htcondor.SecMan()
        status = 0
        message = ""
        _LOG.info("Not verifying that compute resources exist.")
        try:
            for daemon_type in [htcondor.DaemonTypes.Schedd, htcondor.DaemonTypes.Collector]:
                _ = secman.ping(coll.locate(daemon_type))
        except htcondor.HTCondorLocateError:
            status = 1
            message = f"Could not locate {daemon_type} service."
        except htcondor.HTCondorIOError:
            status = 1
            message = f"Permission problem with {daemon_type} service."
        return status, message



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Class and utility functions at the workflow level."""

__all__ = ["HTCondorWorkflow"]


import logging
import os

from lsst.ctrl.bps import (
    BaseWmsWorkflow,
)

from .prepare_utils import _generic_workflow_to_htcondor_dag

_LOG = logging.getLogger(__name__)


class HTCondorWorkflow(BaseWmsWorkflow):
    """Single HTCondor workflow.

    Parameters
    ----------
    name : `str`
        Unique name for Workflow used when naming files.
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration that includes necessary submit/runtime information.
    """

    def __init__(self, name, config=None):
        super().__init__(name, config)
        self.dag = None

    @classmethod
    def from_generic_workflow(cls, config, generic_workflow, out_prefix, service_class):
        # Docstring inherited
        htc_workflow = cls(generic_workflow.name, config)
        htc_workflow.dag = _generic_workflow_to_htcondor_dag(config, generic_workflow, out_prefix)

        _LOG.debug("htcondor dag attribs %s", generic_workflow.run_attrs)
        # Add extra attributes to top most DAG.
        htc_workflow.dag.add_attribs(
            {
                "bps_wms_service": service_class,
                "bps_wms_workflow": f"{cls.__module__}.{cls.__name__}",
            }
        )

        return htc_workflow

    def write(self, out_prefix):
        """Output HTCondor DAGMan files needed for workflow submission.

        Parameters
        ----------
        out_prefix : `str`
            Directory prefix for HTCondor files.
        """
        self.submit_path = out_prefix
        os.makedirs(out_prefix, exist_ok=True)

        # Write down the workflow in HTCondor format.
        self.dag.write(out_prefix, job_subdir="jobs/{self.label}")
